@@ -39,23 +39,22 @@
  */
 
 #define SPICE_CURSOR_CHANNEL_GET_PRIVATE(obj)                                  \
-    (G_TYPE_INSTANCE_GET_PRIVATE((obj), SPICE_TYPE_CURSOR_CHANNEL, spice_cursor_channel))
+    (G_TYPE_INSTANCE_GET_PRIVATE((obj), SPICE_TYPE_CURSOR_CHANNEL, SpiceCursorChannelPrivate))
 
 typedef struct display_cursor display_cursor;
 
 struct display_cursor {
     SpiceCursorHeader           hdr;
     gboolean                    default_cursor;
+    int                         refcount;
     guint32                     data[];
 };
 
-struct spice_cursor_channel {
+struct _SpiceCursorChannelPrivate {
     display_cache               cursors;
     gboolean                    init_done;
 };
 
-G_DEFINE_TYPE(SpiceCursorChannel, spice_cursor_channel, SPICE_TYPE_CHANNEL)
-
 enum {
     SPICE_CURSOR_SET,
     SPICE_CURSOR_MOVE,
@@ -67,14 +66,18 @@ enum {
 
 static guint signals[SPICE_CURSOR_LAST_SIGNAL];
 
-static void spice_cursor_handle_msg(SpiceChannel *channel, spice_msg_in *msg);
+static void spice_cursor_handle_msg(SpiceChannel *channel, SpiceMsgIn *msg);
 static void delete_cursor_all(SpiceChannel *channel);
+static display_cursor * display_cursor_ref(display_cursor *cursor);
+static void display_cursor_unref(display_cursor *cursor);
+
+G_DEFINE_TYPE(SpiceCursorChannel, spice_cursor_channel, SPICE_TYPE_CHANNEL)
 
 /* ------------------------------------------------------------------ */
 
 static void spice_cursor_channel_init(SpiceCursorChannel *channel)
 {
-    spice_cursor_channel *c;
+    SpiceCursorChannelPrivate *c;
 
     c = channel->priv = SPICE_CURSOR_CHANNEL_GET_PRIVATE(channel);
     memset(c, 0, sizeof(*c));
@@ -105,7 +108,9 @@ static void spice_cursor_channel_class_init(SpiceCursorChannelClass *klass)
      * @height: height of the shape
      * @hot_x: horizontal offset of the 'hotspot' of the cursor
      * @hot_y: vertical offset of the 'hotspot' of the cursor
-     * @rgba: 32bits shape data, or %NULL if default cursor
+     * @rgba: 32bits shape data, or %NULL if default cursor. It might
+     * be freed after the signal is emitted, so make sure to copy it
+     * if you need it later!
      *
      * The #SpiceCursorChannel::cursor-set signal is emitted to modify
      * cursor aspect and position on the display area.
@@ -177,7 +182,7 @@ static void spice_cursor_channel_class_init(SpiceCursorChannelClass *klass)
                      G_TYPE_NONE,
                      0);
 
-    g_type_class_add_private(klass, sizeof(spice_cursor_channel));
+    g_type_class_add_private(klass, sizeof(SpiceCursorChannelPrivate));
 }
 
 /* signal trampoline---------------------------------------------------------- */
@@ -290,9 +295,28 @@ static guint32 get_pix_hack(gint pix_index, gint width)
     return (((pix_index % width) ^ (pix_index / width)) & 1) ? 0xc0303030 : 0x30505050;
 }
 
+static display_cursor * display_cursor_ref(display_cursor *cursor)
+{
+    g_return_val_if_fail(cursor != NULL, NULL);
+    g_return_val_if_fail(cursor->refcount > 0, NULL);
+
+    cursor->refcount++;
+    return cursor;
+}
+
+static void display_cursor_unref(display_cursor *cursor)
+{
+    g_return_if_fail(cursor != NULL);
+    g_return_if_fail(cursor->refcount > 0);
+
+    cursor->refcount--;
+    if (cursor->refcount == 0)
+        g_free(cursor);
+}
+
 static display_cursor *set_cursor(SpiceChannel *channel, SpiceCursor *scursor)
 {
-    spice_cursor_channel *c = SPICE_CURSOR_CHANNEL(channel)->priv;
+    SpiceCursorChannelPrivate *c = SPICE_CURSOR_CHANNEL(channel)->priv;
     SpiceCursorHeader *hdr = &scursor->header;
     display_cache_item *item;
     display_cursor *cursor;
@@ -311,10 +335,8 @@ static display_cursor *set_cursor(SpiceChannel *channel, SpiceCursor *scursor)
 
     if (scursor->flags & SPICE_CURSOR_FLAGS_FROM_CACHE) {
         item = cache_find(&c->cursors, hdr->unique);
-        if (!item) {
-            return NULL;
-        }
-        return item->ptr;
+        g_return_val_if_fail(item != NULL, NULL);
+        return display_cursor_ref(item->ptr);
     }
 
     g_return_val_if_fail(scursor->data_size != 0, NULL);
@@ -323,6 +345,7 @@ static display_cursor *set_cursor(SpiceChannel *channel, SpiceCursor *scursor)
     cursor = spice_malloc(sizeof(*cursor) + size);
     cursor->hdr = *hdr;
     cursor->default_cursor = FALSE;
+    cursor->refcount = 1;
     data = scursor->data;
 
     switch (hdr->type) {
@@ -386,6 +409,7 @@ static display_cursor *set_cursor(SpiceChannel *channel, SpiceCursor *scursor)
 
 cache_add:
     if (cursor && (scursor->flags & SPICE_CURSOR_FLAGS_CACHE_ME)) {
+        display_cursor_ref(cursor);
         item = cache_add(&c->cursors, hdr->unique);
         item->ptr = cursor;
     }
@@ -395,15 +419,15 @@ cache_add:
 
 static void delete_cursor_one(SpiceChannel *channel, display_cache_item *item)
 {
-    spice_cursor_channel *c = SPICE_CURSOR_CHANNEL(channel)->priv;
+    SpiceCursorChannelPrivate *c = SPICE_CURSOR_CHANNEL(channel)->priv;
 
-    free(item->ptr);
+    display_cursor_unref((display_cursor*)item->ptr);
     cache_del(&c->cursors, item);
 }
 
 static void delete_cursor_all(SpiceChannel *channel)
 {
-    spice_cursor_channel *c = SPICE_CURSOR_CHANNEL(channel)->priv;
+    SpiceCursorChannelPrivate *c = SPICE_CURSOR_CHANNEL(channel)->priv;
     display_cache_item *item;
 
     for (;;) {
@@ -426,10 +450,10 @@ static void emit_cursor_set(SpiceChannel *channel, display_cursor *cursor)
 }
 
 /* coroutine context */
-static void cursor_handle_init(SpiceChannel *channel, spice_msg_in *in)
+static void cursor_handle_init(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgCursorInit *init = spice_msg_in_parsed(in);
-    spice_cursor_channel *c = SPICE_CURSOR_CHANNEL(channel)->priv;
+    SpiceCursorChannelPrivate *c = SPICE_CURSOR_CHANNEL(channel)->priv;
     display_cursor *cursor;
 
     g_return_if_fail(c->init_done == FALSE);
@@ -437,16 +461,18 @@ static void cursor_handle_init(SpiceChannel *channel, spice_msg_in *in)
     delete_cursor_all(channel);
     cursor = set_cursor(channel, &init->cursor);
     c->init_done = TRUE;
-    if (init->visible && cursor)
+    if (cursor)
         emit_cursor_set(channel, cursor);
-    else
+    if (!init->visible || !cursor)
         emit_main_context(channel, SPICE_CURSOR_HIDE);
+    if (cursor)
+        display_cursor_unref(cursor);
 }
 
 /* coroutine context */
-static void cursor_handle_reset(SpiceChannel *channel, spice_msg_in *in)
+static void cursor_handle_reset(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    spice_cursor_channel *c = SPICE_CURSOR_CHANNEL(channel)->priv;
+    SpiceCursorChannelPrivate *c = SPICE_CURSOR_CHANNEL(channel)->priv;
 
     SPICE_DEBUG("%s, init_done: %d", __FUNCTION__, c->init_done);
 
@@ -456,23 +482,30 @@ static void cursor_handle_reset(SpiceChannel *channel, spice_msg_in *in)
 }
 
 /* coroutine context */
-static void cursor_handle_set(SpiceChannel *channel, spice_msg_in *in)
+static void cursor_handle_set(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgCursorSet *set = spice_msg_in_parsed(in);
-    spice_cursor_channel *c = SPICE_CURSOR_CHANNEL(channel)->priv;
+    SpiceCursorChannelPrivate *c = SPICE_CURSOR_CHANNEL(channel)->priv;
     display_cursor *cursor;
 
     g_return_if_fail(c->init_done == TRUE);
 
     cursor = set_cursor(channel, &set->cursor);
-    emit_cursor_set(channel, cursor);
+    if (cursor)
+        emit_cursor_set(channel, cursor);
+    else
+        emit_main_context(channel, SPICE_CURSOR_HIDE);
+
+
+    if (cursor)
+        display_cursor_unref(cursor);
 }
 
 /* coroutine context */
-static void cursor_handle_move(SpiceChannel *channel, spice_msg_in *in)
+static void cursor_handle_move(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgCursorMove *move = spice_msg_in_parsed(in);
-    spice_cursor_channel *c = SPICE_CURSOR_CHANNEL(channel)->priv;
+    SpiceCursorChannelPrivate *c = SPICE_CURSOR_CHANNEL(channel)->priv;
 
     g_return_if_fail(c->init_done == TRUE);
 
@@ -481,10 +514,10 @@ static void cursor_handle_move(SpiceChannel *channel, spice_msg_in *in)
 }
 
 /* coroutine context */
-static void cursor_handle_hide(SpiceChannel *channel, spice_msg_in *in)
+static void cursor_handle_hide(SpiceChannel *channel, SpiceMsgIn *in)
 {
 #ifdef EXTRA_CHECKS
-    spice_cursor_channel *c = SPICE_CURSOR_CHANNEL(channel)->priv;
+    SpiceCursorChannelPrivate *c = SPICE_CURSOR_CHANNEL(channel)->priv;
 
     g_return_if_fail(c->init_done == TRUE);
 #endif
@@ -493,9 +526,9 @@ static void cursor_handle_hide(SpiceChannel *channel, spice_msg_in *in)
 }
 
 /* coroutine context */
-static void cursor_handle_trail(SpiceChannel *channel, spice_msg_in *in)
+static void cursor_handle_trail(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    spice_cursor_channel *c = SPICE_CURSOR_CHANNEL(channel)->priv;
+    SpiceCursorChannelPrivate *c = SPICE_CURSOR_CHANNEL(channel)->priv;
 
     g_return_if_fail(c->init_done == TRUE);
 
@@ -503,9 +536,9 @@ static void cursor_handle_trail(SpiceChannel *channel, spice_msg_in *in)
 }
 
 /* coroutine context */
-static void cursor_handle_inval_one(SpiceChannel *channel, spice_msg_in *in)
+static void cursor_handle_inval_one(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    spice_cursor_channel *c = SPICE_CURSOR_CHANNEL(channel)->priv;
+    SpiceCursorChannelPrivate *c = SPICE_CURSOR_CHANNEL(channel)->priv;
     SpiceMsgDisplayInvalOne *zap = spice_msg_in_parsed(in);
     display_cache_item *item;
 
@@ -516,19 +549,12 @@ static void cursor_handle_inval_one(SpiceChannel *channel, spice_msg_in *in)
 }
 
 /* coroutine context */
-static void cursor_handle_inval_all(SpiceChannel *channel, spice_msg_in *in)
+static void cursor_handle_inval_all(SpiceChannel *channel, SpiceMsgIn *in)
 {
     delete_cursor_all(channel);
 }
 
-static spice_msg_handler cursor_handlers[] = {
-    [ SPICE_MSG_SET_ACK ]                  = spice_channel_handle_set_ack,
-    [ SPICE_MSG_PING ]                     = spice_channel_handle_ping,
-    [ SPICE_MSG_NOTIFY ]                   = spice_channel_handle_notify,
-    [ SPICE_MSG_DISCONNECTING ]            = spice_channel_handle_disconnect,
-    [ SPICE_MSG_WAIT_FOR_CHANNELS ]        = spice_channel_handle_wait_for_channels,
-    [ SPICE_MSG_MIGRATE ]                  = spice_channel_handle_migrate,
-
+static const spice_msg_handler cursor_handlers[] = {
     [ SPICE_MSG_CURSOR_INIT ]              = cursor_handle_init,
     [ SPICE_MSG_CURSOR_RESET ]             = cursor_handle_reset,
     [ SPICE_MSG_CURSOR_SET ]               = cursor_handle_set,
@@ -540,10 +566,19 @@ static spice_msg_handler cursor_handlers[] = {
 };
 
 /* coroutine context */
-static void spice_cursor_handle_msg(SpiceChannel *channel, spice_msg_in *msg)
+static void spice_cursor_handle_msg(SpiceChannel *channel, SpiceMsgIn *msg)
 {
     int type = spice_msg_in_type(msg);
+    SpiceChannelClass *parent_class;
+
     g_return_if_fail(type < SPICE_N_ELEMENTS(cursor_handlers));
-    g_return_if_fail(cursor_handlers[type] != NULL);
-    cursor_handlers[type](channel, msg);
+
+    parent_class = SPICE_CHANNEL_CLASS(spice_cursor_channel_parent_class);
+
+    if (cursor_handlers[type] != NULL)
+        cursor_handlers[type](channel, msg);
+    else if (parent_class->handle_msg)
+        parent_class->handle_msg(channel, msg);
+    else
+        g_return_if_reached();
 }
