@@ -31,7 +31,6 @@
 #include "spice-common.h"
 
 #include "spice-marshal.h"
-#include "spice-channel-cache.h"
 #include "spice-channel-priv.h"
 #include "spice-session-priv.h"
 #include "channel-display-priv.h"
@@ -57,12 +56,12 @@
  */
 
 #define SPICE_DISPLAY_CHANNEL_GET_PRIVATE(obj)                                  \
-    (G_TYPE_INSTANCE_GET_PRIVATE((obj), SPICE_TYPE_DISPLAY_CHANNEL, spice_display_channel))
+    (G_TYPE_INSTANCE_GET_PRIVATE((obj), SPICE_TYPE_DISPLAY_CHANNEL, SpiceDisplayChannelPrivate))
 
-struct spice_display_channel {
+struct _SpiceDisplayChannelPrivate {
     Ring                        surfaces;
-    display_cache               images;
-    display_cache               palettes;
+    display_cache               *images;
+    display_cache               *palettes;
     SpiceImageCache             image_cache;
     SpicePaletteCache           palette_cache;
     SpiceImageSurfaces          image_surfaces;
@@ -70,6 +69,7 @@ struct spice_display_channel {
     display_stream              **streams;
     int                         nstreams;
     gboolean                    mark;
+    guint                       mark_false_event_id;
 #ifdef WIN32
     HDC dc;
 #endif
@@ -77,6 +77,13 @@ struct spice_display_channel {
 
 G_DEFINE_TYPE(SpiceDisplayChannel, spice_display_channel, SPICE_TYPE_CHANNEL)
 
+/* Properties */
+enum {
+    PROP_0,
+    PROP_WIDTH,
+    PROP_HEIGHT,
+};
+
 enum {
     SPICE_DISPLAY_PRIMARY_CREATE,
     SPICE_DISPLAY_PRIMARY_DESTROY,
@@ -88,42 +95,124 @@ enum {
 
 static guint signals[SPICE_DISPLAY_LAST_SIGNAL];
 
-static void spice_display_handle_msg(SpiceChannel *channel, spice_msg_in *msg);
+static void spice_display_handle_msg(SpiceChannel *channel, SpiceMsgIn *msg);
 static void spice_display_channel_init(SpiceDisplayChannel *channel);
 static void spice_display_channel_up(SpiceChannel *channel);
 
-static void palette_clear(SpicePaletteCache *cache);
-static void image_clear(SpiceImageCache *cache);
 static void clear_surfaces(SpiceChannel *channel);
 static void clear_streams(SpiceChannel *channel);
-static display_surface *find_surface(spice_display_channel *c, int surface_id);
+static display_surface *find_surface(SpiceDisplayChannelPrivate *c, int surface_id);
 static gboolean display_stream_render(display_stream *st);
 
 /* ------------------------------------------------------------------ */
 
+static void spice_display_channel_dispose(GObject *object)
+{
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(object)->priv;
+
+    if (c->mark_false_event_id != 0) {
+        g_source_remove(c->mark_false_event_id);
+        c->mark_false_event_id = 0;
+    }
+
+    if (G_OBJECT_CLASS(spice_display_channel_parent_class)->dispose)
+        G_OBJECT_CLASS(spice_display_channel_parent_class)->dispose(object);
+}
+
 static void spice_display_channel_finalize(GObject *obj)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(obj)->priv;
-
-    palette_clear(&c->palette_cache);
-    image_clear(&c->image_cache);
     clear_surfaces(SPICE_CHANNEL(obj));
     clear_streams(SPICE_CHANNEL(obj));
-    glz_decoder_window_destroy(c->glz_window);
 
     if (G_OBJECT_CLASS(spice_display_channel_parent_class)->finalize)
         G_OBJECT_CLASS(spice_display_channel_parent_class)->finalize(obj);
 }
 
+static void spice_display_channel_constructed(GObject *object)
+{
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(object)->priv;
+    SpiceSession *s = spice_channel_get_session(SPICE_CHANNEL(object));
+
+    g_return_if_fail(s != NULL);
+    spice_session_get_caches(s, &c->images, &c->palettes, &c->glz_window);
+
+    g_return_if_fail(c->glz_window != NULL);
+    g_return_if_fail(c->images != NULL);
+    g_return_if_fail(c->palettes != NULL);
+
+    if (G_OBJECT_CLASS(spice_display_channel_parent_class)->constructed)
+        G_OBJECT_CLASS(spice_display_channel_parent_class)->constructed(object);
+}
+
+
+static void spice_display_get_property(GObject    *object,
+                                       guint       prop_id,
+                                       GValue     *value,
+                                       GParamSpec *pspec)
+{
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(object)->priv;
+
+    switch (prop_id) {
+    case PROP_WIDTH: {
+        display_surface *surface = find_surface(c, 0);
+        g_value_set_uint(value, surface ? surface->width : 0);
+        break;
+    }
+    case PROP_HEIGHT: {
+        display_surface *surface = find_surface(c, 0);
+        g_value_set_uint(value, surface ? surface->height : 0);
+        break;
+    }
+    default:
+        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
+        break;
+    }
+}
+
+static void spice_display_set_property(GObject      *object,
+                                       guint         prop_id,
+                                       const GValue *value,
+                                       GParamSpec   *pspec)
+{
+    switch (prop_id) {
+    default:
+        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
+        break;
+    }
+}
+
 static void spice_display_channel_class_init(SpiceDisplayChannelClass *klass)
 {
     GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
     SpiceChannelClass *channel_class = SPICE_CHANNEL_CLASS(klass);
 
     gobject_class->finalize     = spice_display_channel_finalize;
+    gobject_class->dispose      = spice_display_channel_dispose;
+    gobject_class->get_property = spice_display_get_property;
+    gobject_class->set_property = spice_display_set_property;
+    gobject_class->constructed = spice_display_channel_constructed;
+
     channel_class->handle_msg   = spice_display_handle_msg;
     channel_class->channel_up   = spice_display_channel_up;
 
+    g_object_class_install_property
+        (gobject_class, PROP_HEIGHT,
+         g_param_spec_uint("height",
+                           "Display height",
+                           "The primary surface height",
+                           0, G_MAXUINT, 0,
+                           G_PARAM_READABLE |
+                           G_PARAM_STATIC_STRINGS));
+
+    g_object_class_install_property
+        (gobject_class, PROP_WIDTH,
+         g_param_spec_uint("width",
+                           "Display width",
+                           "The primary surface width",
+                           0, G_MAXUINT, 0,
+                           G_PARAM_READABLE |
+                           G_PARAM_STATIC_STRINGS));
+
     /**
      * SpiceDisplayChannel::display-primary-create:
      * @display: the #SpiceDisplayChannel that emitted the signal
@@ -214,7 +303,7 @@ static void spice_display_channel_class_init(SpiceDisplayChannelClass *klass)
                      1,
                      G_TYPE_INT);
 
-    g_type_class_add_private(klass, sizeof(spice_display_channel));
+    g_type_class_add_private(klass, sizeof(SpiceDisplayChannelPrivate));
 
     sw_canvas_init();
     quic_init();
@@ -279,29 +368,29 @@ static void do_emit_main_context(GObject *object, int signum, gpointer params)
 
 static void image_put(SpiceImageCache *cache, uint64_t id, pixman_image_t *image)
 {
-    spice_display_channel *c =
-        SPICE_CONTAINEROF(cache, spice_display_channel, image_cache);
+    SpiceDisplayChannelPrivate *c =
+        SPICE_CONTAINEROF(cache, SpiceDisplayChannelPrivate, image_cache);
     display_cache_item *item;
 
-    item = cache_find(&c->images, id);
+    item = cache_find(c->images, id);
     if (item) {
         cache_ref(item);
         return;
     }
 
-    item = cache_add(&c->images, id);
+    item = cache_add(c->images, id);
     item->ptr = pixman_image_ref(image);
 }
 
 static pixman_image_t *image_get(SpiceImageCache *cache, uint64_t id)
 {
-    spice_display_channel *c =
-        SPICE_CONTAINEROF(cache, spice_display_channel, image_cache);
+    SpiceDisplayChannelPrivate *c =
+        SPICE_CONTAINEROF(cache, SpiceDisplayChannelPrivate, image_cache);
     display_cache_item *item;
 
-    item = cache_find(&c->images, id);
+    item = cache_find(c->images, id);
     if (item) {
-        cache_used(&c->images, item);
+        cache_used(c->images, item);
         return pixman_image_ref(item->ptr);
     }
     return NULL;
@@ -309,52 +398,36 @@ static pixman_image_t *image_get(SpiceImageCache *cache, uint64_t id)
 
 static void image_remove(SpiceImageCache *cache, uint64_t id)
 {
-    spice_display_channel *c =
-        SPICE_CONTAINEROF(cache, spice_display_channel, image_cache);
+    SpiceDisplayChannelPrivate *c =
+        SPICE_CONTAINEROF(cache, SpiceDisplayChannelPrivate, image_cache);
     display_cache_item *item;
 
-    item = cache_find(&c->images, id);
+    item = cache_find(c->images, id);
     g_return_if_fail(item != NULL);
     if (cache_unref(item)) {
         pixman_image_unref(item->ptr);
-        cache_del(&c->images, item);
-    }
-}
-
-static void image_clear(SpiceImageCache *cache)
-{
-    spice_display_channel *c =
-        SPICE_CONTAINEROF(cache, spice_display_channel, image_cache);
-    display_cache_item *item;
-
-    for (;;) {
-        item = cache_get_lru(&c->images);
-        if (item == NULL) {
-            break;
-        }
-        pixman_image_unref(item->ptr);
-        cache_del(&c->images, item);
+        cache_del(c->images, item);
     }
 }
 
 static void palette_put(SpicePaletteCache *cache, SpicePalette *palette)
 {
-    spice_display_channel *c =
-        SPICE_CONTAINEROF(cache, spice_display_channel, palette_cache);
+    SpiceDisplayChannelPrivate *c =
+        SPICE_CONTAINEROF(cache, SpiceDisplayChannelPrivate, palette_cache);
     display_cache_item *item;
 
-    item = cache_add(&c->palettes, palette->unique);
+    item = cache_add(c->palettes, palette->unique);
     item->ptr = g_memdup(palette, sizeof(SpicePalette) +
                          palette->num_ents * sizeof(palette->ents[0]));
 }
 
 static SpicePalette *palette_get(SpicePaletteCache *cache, uint64_t id)
 {
-    spice_display_channel *c =
-        SPICE_CONTAINEROF(cache, spice_display_channel, palette_cache);
+    SpiceDisplayChannelPrivate *c =
+        SPICE_CONTAINEROF(cache, SpiceDisplayChannelPrivate, palette_cache);
     display_cache_item *item;
 
-    item = cache_find(&c->palettes, id);
+    item = cache_find(c->palettes, id);
     if (item) {
         cache_ref(item);
         return item->ptr;
@@ -364,34 +437,19 @@ static SpicePalette *palette_get(SpicePaletteCache *cache, uint64_t id)
 
 static void palette_remove(SpicePaletteCache *cache, uint32_t id)
 {
-    spice_display_channel *c =
-        SPICE_CONTAINEROF(cache, spice_display_channel, palette_cache);
+    SpiceDisplayChannelPrivate *c =
+        SPICE_CONTAINEROF(cache, SpiceDisplayChannelPrivate, palette_cache);
     display_cache_item *item;
 
-    item = cache_find(&c->palettes, id);
+    item = cache_find(c->palettes, id);
     if (item) {
         if (cache_unref(item)) {
             g_free(item->ptr);
-            cache_del(&c->palettes, item);
+            cache_del(c->palettes, item);
         }
     }
 }
 
-static void palette_clear(SpicePaletteCache *cache)
-{
-    spice_display_channel *c =
-        SPICE_CONTAINEROF(cache, spice_display_channel, palette_cache);
-    display_cache_item *item;
-
-    for (;;) {
-        item = cache_get_lru(&c->palettes);
-        if (item == NULL) {
-            break;
-        }
-        cache_del(&c->palettes, item);
-    }
-}
-
 static void palette_release(SpicePaletteCache *cache, SpicePalette *palette)
 {
     palette_remove(cache, palette->unique);
@@ -401,15 +459,15 @@ static void palette_release(SpicePaletteCache *cache, SpicePalette *palette)
 static void image_put_lossy(SpiceImageCache *cache, uint64_t id,
                             pixman_image_t *surface)
 {
-    spice_display_channel *c =
-        SPICE_CONTAINEROF(cache, spice_display_channel, image_cache);
+    SpiceDisplayChannelPrivate *c =
+        SPICE_CONTAINEROF(cache, SpiceDisplayChannelPrivate, image_cache);
     display_cache_item *item;
 
-#if 1 /* TODO: temporary sanity check */
-    g_warn_if_fail(cache_find(&c->images, id) == NULL);
+#ifndef NDEBUG
+    g_warn_if_fail(cache_find(c->images, id) == NULL);
 #endif
 
-    item = cache_add(&c->images, id);
+    item = cache_add(c->images, id);
     item->ptr = pixman_image_ref(surface);
     item->lossy = TRUE;
 }
@@ -417,11 +475,11 @@ static void image_put_lossy(SpiceImageCache *cache, uint64_t id,
 static void image_replace_lossy(SpiceImageCache *cache, uint64_t id,
                                 pixman_image_t *surface)
 {
-    spice_display_channel *c =
-        SPICE_CONTAINEROF(cache, spice_display_channel, image_cache);
+    SpiceDisplayChannelPrivate *c =
+        SPICE_CONTAINEROF(cache, SpiceDisplayChannelPrivate, image_cache);
     display_cache_item *item;
 
-    item = cache_find(&c->images, id);
+    item = cache_find(c->images, id);
     g_return_if_fail(item != NULL);
 
     pixman_image_unref(item->ptr);
@@ -431,11 +489,11 @@ static void image_replace_lossy(SpiceImageCache *cache, uint64_t id,
 
 static pixman_image_t* image_get_lossless(SpiceImageCache *cache, uint64_t id)
 {
-    spice_display_channel *c =
-        SPICE_CONTAINEROF(cache, spice_display_channel, image_cache);
+    SpiceDisplayChannelPrivate *c =
+        SPICE_CONTAINEROF(cache, SpiceDisplayChannelPrivate, image_cache);
     display_cache_item *item;
 
-    item = cache_find(&c->images, id);
+    item = cache_find(c->images, id);
     if (!item)
         return NULL;
 
@@ -443,7 +501,7 @@ static pixman_image_t* image_get_lossless(SpiceImageCache *cache, uint64_t id)
        that necessary? */
     g_warn_if_fail(item->lossy == FALSE);
 
-    cache_used(&c->images, item);
+    cache_used(c->images, item);
     return pixman_image_ref(item->ptr);
 }
 #endif
@@ -451,8 +509,8 @@ static pixman_image_t* image_get_lossless(SpiceImageCache *cache, uint64_t id)
 SpiceCanvas *surfaces_get(SpiceImageSurfaces *surfaces,
                           uint32_t surface_id)
 {
-    spice_display_channel *c =
-        SPICE_CONTAINEROF(surfaces, spice_display_channel, image_surfaces);
+    SpiceDisplayChannelPrivate *c =
+        SPICE_CONTAINEROF(surfaces, SpiceDisplayChannelPrivate, image_surfaces);
 
     display_surface *s =
         find_surface(c, surface_id);
@@ -494,14 +552,12 @@ static HDC create_compatible_dc(void)
 
 static void spice_display_channel_init(SpiceDisplayChannel *channel)
 {
-    spice_display_channel *c;
+    SpiceDisplayChannelPrivate *c;
 
     c = channel->priv = SPICE_DISPLAY_CHANNEL_GET_PRIVATE(channel);
     memset(c, 0, sizeof(*c));
 
     ring_init(&c->surfaces);
-    cache_init(&c->images, "image");
-    cache_init(&c->palettes, "palette");
     c->image_cache.ops = &image_cache_ops;
     c->palette_cache.ops = &palette_cache_ops;
     c->image_surfaces.ops = &image_surfaces_ops;
@@ -514,7 +570,7 @@ static void spice_display_channel_init(SpiceDisplayChannel *channel)
 
 static int create_canvas(SpiceChannel *channel, display_surface *surface)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
 
     if (surface->primary) {
         SPICE_DEBUG("display: create primary canvas");
@@ -538,9 +594,7 @@ static int create_canvas(SpiceChannel *channel, display_surface *surface)
         surface->data = spice_malloc(surface->size);
     }
 
-    if (!c->glz_window) {
-        c->glz_window = glz_decoder_window_new();
-    }
+    g_return_val_if_fail(c->glz_window, 0);
 
     g_warn_if_fail(surface->canvas == NULL);
     g_warn_if_fail(surface->glz_decoder == NULL);
@@ -593,7 +647,7 @@ static void destroy_canvas(display_surface *surface)
     surface->canvas = NULL;
 }
 
-static display_surface *find_surface(spice_display_channel *c, int surface_id)
+static display_surface *find_surface(SpiceDisplayChannelPrivate *c, int surface_id)
 {
     display_surface *surface;
     RingItem *item;
@@ -610,7 +664,7 @@ static display_surface *find_surface(spice_display_channel *c, int surface_id)
 
 static void clear_surfaces(SpiceChannel *channel)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     display_surface *surface;
     RingItem *item;
 
@@ -626,15 +680,6 @@ static void clear_surfaces(SpiceChannel *channel)
 /* coroutine context */
 static void emit_invalidate(SpiceChannel *channel, SpiceRect *bbox)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
-
-    /* FIXME: we shouldn't invalidate before the mark is sent, but
-       server-side is not correct in this regard... */
-    if (!c->mark) {
-        c->mark = TRUE;
-        emit_main_context(channel, SPICE_DISPLAY_MARK, TRUE);
-    }
-
     emit_main_context(channel, SPICE_DISPLAY_INVALIDATE,
                       bbox->left, bbox->top,
                       bbox->right - bbox->left,
@@ -646,7 +691,7 @@ static void emit_invalidate(SpiceChannel *channel, SpiceRect *bbox)
 /* coroutine context */
 static void spice_display_channel_up(SpiceChannel *channel)
 {
-    spice_msg_out *out;
+    SpiceMsgOut *out;
     SpiceMsgcDisplayInit init = {
         .pixmap_cache_id            = 1,
         .pixmap_cache_size          = DISPLAY_PIXMAP_CACHE,
@@ -673,9 +718,9 @@ static void spice_display_channel_up(SpiceChannel *channel)
 }
 
 /* coroutine context */
-static void display_handle_mode(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_mode(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     SpiceMsgDisplayMode *mode = spice_msg_in_parsed(in);
     display_surface *surface = find_surface(c, 0);
 
@@ -709,9 +754,9 @@ static void display_handle_mode(SpiceChannel *channel, spice_msg_in *in)
 }
 
 /* coroutine context */
-static void display_handle_mark(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_mark(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     display_surface *surface = find_surface(c, 0);
 
     SPICE_DEBUG("%s", __FUNCTION__);
@@ -725,9 +770,9 @@ static void display_handle_mark(SpiceChannel *channel, spice_msg_in *in)
 }
 
 /* coroutine context */
-static void display_handle_reset(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_reset(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     display_surface *surface = find_surface(c, 0);
 
     SPICE_DEBUG("%s: TODO detach_from_screen", __FUNCTION__);
@@ -735,17 +780,17 @@ static void display_handle_reset(SpiceChannel *channel, spice_msg_in *in)
     if (surface != NULL)
         surface->canvas->ops->clear(surface->canvas);
 
-    palette_clear(&c->palette_cache);
+    spice_session_palettes_clear(spice_channel_get_session(channel));
 
     c->mark = FALSE;
     emit_main_context(channel, SPICE_DISPLAY_MARK, FALSE);
 }
 
 /* coroutine context */
-static void display_handle_copy_bits(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_copy_bits(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgDisplayCopyBits *op = spice_msg_in_parsed(in);
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     display_surface *surface = find_surface(c, op->base.surface_id);
 
     g_return_if_fail(surface != NULL);
@@ -757,9 +802,9 @@ static void display_handle_copy_bits(SpiceChannel *channel, spice_msg_in *in)
 }
 
 /* coroutine context */
-static void display_handle_inv_list(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_inv_list(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     SpiceResourceList *list = spice_msg_in_parsed(in);
     int i;
 
@@ -776,28 +821,24 @@ static void display_handle_inv_list(SpiceChannel *channel, spice_msg_in *in)
 }
 
 /* coroutine context */
-static void display_handle_inv_pixmap_all(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_inv_pixmap_all(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
-
-    image_clear(&c->image_cache);
+    spice_session_images_clear(spice_channel_get_session(channel));
 }
 
 /* coroutine context */
-static void display_handle_inv_palette(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_inv_palette(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     SpiceMsgDisplayInvalOne* op = spice_msg_in_parsed(in);
 
     palette_remove(&c->palette_cache, op->id);
 }
 
 /* coroutine context */
-static void display_handle_inv_palette_all(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_inv_palette_all(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
-
-    palette_clear(&c->palette_cache);
+    spice_session_palettes_clear(spice_channel_get_session(channel));
 }
 
 /* ------------------------------------------------------------------ */
@@ -822,13 +863,13 @@ static void display_update_stream_region(display_stream *st)
 }
 
 /* coroutine context */
-static void display_handle_stream_create(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_stream_create(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     SpiceMsgDisplayStreamCreate *op = spice_msg_in_parsed(in);
     display_stream *st;
 
-    g_message("%s: id %d", __FUNCTION__, op->id);
+    SPICE_DEBUG("%s: id %d", __FUNCTION__, op->id);
 
     if (op->id >= c->nstreams) {
         int n = c->nstreams;
@@ -868,7 +909,7 @@ static gboolean display_stream_schedule(display_stream *st)
 {
     guint32 time, d;
     SpiceMsgDisplayStreamData *op;
-    spice_msg_in *in;
+    SpiceMsgIn *in;
 
     if (st->timeout)
         return TRUE;
@@ -897,7 +938,7 @@ static gboolean display_stream_schedule(display_stream *st)
 /* main context */
 static gboolean display_stream_render(display_stream *st)
 {
-    spice_msg_in *in;
+    SpiceMsgIn *in;
 
     st->timeout = 0;
     do {
@@ -955,16 +996,23 @@ static gboolean display_stream_render(display_stream *st)
 }
 
 /* coroutine context */
-static void display_handle_stream_data(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_stream_data(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     SpiceMsgDisplayStreamData *op = spice_msg_in_parsed(in);
     display_stream *st = c->streams[op->id];
-    guint32 time;
+    guint32 mmtime;
 
-    time = spice_session_get_mm_time(spice_channel_get_session(channel));
-    if (op->multi_media_time < time) {
-        SPICE_DEBUG("stream data too late by %u ms, dropin", time - op->multi_media_time);
+    mmtime = spice_session_get_mm_time(spice_channel_get_session(channel));
+
+    if (op->multi_media_time == 0) {
+        g_critical("Received frame with invalid 0 timestamp! perhaps wrong graphic driver?");
+        op->multi_media_time = mmtime + 100; /* workaround... */
+    }
+
+    if (op->multi_media_time < mmtime) {
+        SPICE_DEBUG("stream data too late by %u ms (ts: %u, mmtime: %u), dropin",
+                    mmtime - op->multi_media_time, op->multi_media_time, mmtime);
         return;
     }
 
@@ -974,9 +1022,9 @@ static void display_handle_stream_data(SpiceChannel *channel, spice_msg_in *in)
 }
 
 /* coroutine context */
-static void display_handle_stream_clip(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_stream_clip(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     SpiceMsgDisplayStreamClip *op = spice_msg_in_parsed(in);
     display_stream *st = c->streams[op->id];
 
@@ -996,10 +1044,11 @@ static void _msg_in_unref_func(gpointer data, gpointer user_data)
 
 static void destroy_stream(SpiceChannel *channel, int id)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     display_stream *st;
 
     g_return_if_fail(c != NULL);
+    g_return_if_fail(c->streams != NULL);
 
     st = c->streams[id];
     if (!st)
@@ -1025,7 +1074,7 @@ static void destroy_stream(SpiceChannel *channel, int id)
 
 static void clear_streams(SpiceChannel *channel)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     int i;
 
     for (i = 0; i < c->nstreams; i++) {
@@ -1037,17 +1086,17 @@ static void clear_streams(SpiceChannel *channel)
 }
 
 /* coroutine context */
-static void display_handle_stream_destroy(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_stream_destroy(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgDisplayStreamDestroy *op = spice_msg_in_parsed(in);
 
     g_return_if_fail(op != NULL);
-    g_message("%s: id %d", __FUNCTION__, op->id);
+    SPICE_DEBUG("%s: id %d", __FUNCTION__, op->id);
     destroy_stream(channel, op->id);
 }
 
 /* coroutine context */
-static void display_handle_stream_destroy_all(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_stream_destroy_all(SpiceChannel *channel, SpiceMsgIn *in)
 {
     clear_streams(channel);
 }
@@ -1055,92 +1104,92 @@ static void display_handle_stream_destroy_all(SpiceChannel *channel, spice_msg_i
 /* ------------------------------------------------------------------ */
 
 /* coroutine context */
-static void display_handle_draw_fill(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_draw_fill(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgDisplayDrawFill *op = spice_msg_in_parsed(in);
     DRAW(fill);
 }
 
 /* coroutine context */
-static void display_handle_draw_opaque(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_draw_opaque(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgDisplayDrawOpaque *op = spice_msg_in_parsed(in);
     DRAW(opaque);
 }
 
 /* coroutine context */
-static void display_handle_draw_copy(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_draw_copy(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgDisplayDrawCopy *op = spice_msg_in_parsed(in);
     DRAW(copy);
 }
 
 /* coroutine context */
-static void display_handle_draw_blend(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_draw_blend(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgDisplayDrawBlend *op = spice_msg_in_parsed(in);
     DRAW(blend);
 }
 
 /* coroutine context */
-static void display_handle_draw_blackness(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_draw_blackness(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgDisplayDrawBlackness *op = spice_msg_in_parsed(in);
     DRAW(blackness);
 }
 
-static void display_handle_draw_whiteness(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_draw_whiteness(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgDisplayDrawWhiteness *op = spice_msg_in_parsed(in);
     DRAW(whiteness);
 }
 
 /* coroutine context */
-static void display_handle_draw_invers(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_draw_invers(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgDisplayDrawInvers *op = spice_msg_in_parsed(in);
     DRAW(invers);
 }
 
 /* coroutine context */
-static void display_handle_draw_rop3(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_draw_rop3(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgDisplayDrawRop3 *op = spice_msg_in_parsed(in);
     DRAW(rop3);
 }
 
 /* coroutine context */
-static void display_handle_draw_stroke(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_draw_stroke(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgDisplayDrawStroke *op = spice_msg_in_parsed(in);
     DRAW(stroke);
 }
 
 /* coroutine context */
-static void display_handle_draw_text(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_draw_text(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgDisplayDrawText *op = spice_msg_in_parsed(in);
     DRAW(text);
 }
 
 /* coroutine context */
-static void display_handle_draw_transparent(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_draw_transparent(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgDisplayDrawTransparent *op = spice_msg_in_parsed(in);
     DRAW(transparent);
 }
 
 /* coroutine context */
-static void display_handle_draw_alpha_blend(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_draw_alpha_blend(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgDisplayDrawAlphaBlend *op = spice_msg_in_parsed(in);
     DRAW(alpha_blend);
 }
 
 /* coroutine context */
-static void display_handle_surface_create(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_surface_create(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     SpiceMsgSurfaceCreate *create = spice_msg_in_parsed(in);
     display_surface *surface = spice_new0(display_surface, 1);
 
@@ -1157,6 +1206,10 @@ static void display_handle_surface_create(SpiceChannel *channel, spice_msg_in *i
         emit_main_context(channel, SPICE_DISPLAY_PRIMARY_CREATE,
                           surface->format, surface->width, surface->height,
                           surface->stride, surface->shmid, surface->data);
+        if (c->mark_false_event_id != 0) {
+            g_source_remove(c->mark_false_event_id);
+            c->mark_false_event_id = FALSE;
+        }
     } else {
         surface->primary = false;
         create_canvas(channel, surface);
@@ -1165,11 +1218,23 @@ static void display_handle_surface_create(SpiceChannel *channel, spice_msg_in *i
     ring_add(&c->surfaces, &surface->link);
 }
 
+static gboolean display_mark_false(gpointer data)
+{
+    SpiceChannel *channel = data;
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+
+    c->mark = FALSE;
+    g_signal_emit(channel, signals[SPICE_DISPLAY_MARK], 0, FALSE);
+
+    c->mark_false_event_id = 0;
+    return FALSE;
+}
+
 /* coroutine context */
-static void display_handle_surface_destroy(SpiceChannel *channel, spice_msg_in *in)
+static void display_handle_surface_destroy(SpiceChannel *channel, SpiceMsgIn *in)
 {
     SpiceMsgSurfaceDestroy *destroy = spice_msg_in_parsed(in);
-    spice_display_channel *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
+    SpiceDisplayChannelPrivate *c = SPICE_DISPLAY_CHANNEL(channel)->priv;
     display_surface *surface;
 
     g_return_if_fail(destroy != NULL);
@@ -1181,6 +1246,12 @@ static void display_handle_surface_destroy(SpiceChannel *channel, spice_msg_in *
         return;
     }
     if (surface->primary) {
+        int id = spice_channel_get_channel_id(channel);
+        SPICE_DEBUG("%d: FIXME primary destroy, but is display really disabled?", id);
+        /* this is done with a timeout in spicec as well, it's *ugly* */
+        if (id != 0 && c->mark_false_event_id == 0) {
+            c->mark_false_event_id = g_timeout_add_seconds(1, display_mark_false, channel);
+        }
         emit_main_context(channel, SPICE_DISPLAY_PRIMARY_DESTROY);
     }
 
@@ -1189,14 +1260,7 @@ static void display_handle_surface_destroy(SpiceChannel *channel, spice_msg_in *
     free(surface);
 }
 
-static spice_msg_handler display_handlers[] = {
-    [ SPICE_MSG_SET_ACK ]                    = spice_channel_handle_set_ack,
-    [ SPICE_MSG_PING ]                       = spice_channel_handle_ping,
-    [ SPICE_MSG_NOTIFY ]                     = spice_channel_handle_notify,
-    [ SPICE_MSG_DISCONNECTING ]              = spice_channel_handle_disconnect,
-    [ SPICE_MSG_WAIT_FOR_CHANNELS ]          = spice_channel_handle_wait_for_channels,
-    [ SPICE_MSG_MIGRATE ]                    = spice_channel_handle_migrate,
-
+static const spice_msg_handler display_handlers[] = {
     [ SPICE_MSG_DISPLAY_MODE ]               = display_handle_mode,
     [ SPICE_MSG_DISPLAY_MARK ]               = display_handle_mark,
     [ SPICE_MSG_DISPLAY_RESET ]              = display_handle_reset,
@@ -1230,10 +1294,19 @@ static spice_msg_handler display_handlers[] = {
 };
 
 /* coroutine context */
-static void spice_display_handle_msg(SpiceChannel *channel, spice_msg_in *msg)
+static void spice_display_handle_msg(SpiceChannel *channel, SpiceMsgIn *msg)
 {
     int type = spice_msg_in_type(msg);
+    SpiceChannelClass *parent_class;
+
     g_return_if_fail(type < SPICE_N_ELEMENTS(display_handlers));
-    g_return_if_fail(display_handlers[type] != NULL);
-    display_handlers[type](channel, msg);
+
+    parent_class = SPICE_CHANNEL_CLASS(spice_display_channel_parent_class);
+
+    if (display_handlers[type] != NULL)
+        display_handlers[type](channel, msg);
+    else if (parent_class->handle_msg)
+        parent_class->handle_msg(channel, msg);
+    else
+        g_return_if_reached();
 }
