@@ -46,9 +46,9 @@
  */
 
 #define SPICE_RECORD_CHANNEL_GET_PRIVATE(obj)                                  \
-    (G_TYPE_INSTANCE_GET_PRIVATE((obj), SPICE_TYPE_RECORD_CHANNEL, spice_record_channel))
+    (G_TYPE_INSTANCE_GET_PRIVATE((obj), SPICE_TYPE_RECORD_CHANNEL, SpiceRecordChannelPrivate))
 
-struct spice_record_channel {
+struct _SpiceRecordChannelPrivate {
     int                         mode;
     gboolean                    started;
     CELTMode                    *celt_mode;
@@ -56,10 +56,22 @@ struct spice_record_channel {
     gsize                       frame_bytes;
     guint8                      *last_frame;
     gsize                       last_frame_current;
+    guint8                      nchannels;
+    guint16                     *volume;
+    guint8                      mute;
 };
 
 G_DEFINE_TYPE(SpiceRecordChannel, spice_record_channel, SPICE_TYPE_CHANNEL)
 
+/* Properties */
+enum {
+    PROP_0,
+    PROP_NCHANNELS,
+    PROP_VOLUME,
+    PROP_MUTE,
+};
+
+/* Signals */
 enum {
     SPICE_RECORD_START,
     SPICE_RECORD_STOP,
@@ -69,7 +81,7 @@ enum {
 
 static guint signals[SPICE_RECORD_LAST_SIGNAL];
 
-static void spice_record_handle_msg(SpiceChannel *channel, spice_msg_in *msg);
+static void spice_record_handle_msg(SpiceChannel *channel, SpiceMsgIn *msg);
 static void channel_up(SpiceChannel *channel);
 
 #define FRAME_SIZE 256
@@ -79,16 +91,17 @@ static void channel_up(SpiceChannel *channel);
 
 static void spice_record_channel_init(SpiceRecordChannel *channel)
 {
-    spice_record_channel *c;
+    SpiceRecordChannelPrivate *c;
 
     c = channel->priv = SPICE_RECORD_CHANNEL_GET_PRIVATE(channel);
     memset(c, 0, sizeof(*c));
     spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_RECORD_CAP_CELT_0_5_1);
+    spice_channel_set_capability(SPICE_CHANNEL(channel), SPICE_RECORD_CAP_VOLUME);
 }
 
 static void spice_record_channel_finalize(GObject *obj)
 {
-    spice_record_channel *c = SPICE_RECORD_CHANNEL(obj)->priv;
+    SpiceRecordChannelPrivate *c = SPICE_RECORD_CHANNEL(obj)->priv;
 
     g_free(c->last_frame);
     c->last_frame = NULL;
@@ -103,19 +116,91 @@ static void spice_record_channel_finalize(GObject *obj)
         c->celt_mode = NULL;
     }
 
+    g_free(c->volume);
+    c->volume = NULL;
+
     if (G_OBJECT_CLASS(spice_record_channel_parent_class)->finalize)
         G_OBJECT_CLASS(spice_record_channel_parent_class)->finalize(obj);
 }
 
+static void spice_record_channel_get_property(GObject    *gobject,
+                                              guint       prop_id,
+                                              GValue     *value,
+                                              GParamSpec *pspec)
+{
+    SpiceRecordChannel *channel = SPICE_RECORD_CHANNEL(gobject);
+    SpiceRecordChannelPrivate *c = channel->priv;
+
+    switch (prop_id) {
+    case PROP_VOLUME:
+        g_value_set_pointer(value, c->volume);
+        break;
+    case PROP_NCHANNELS:
+        g_value_set_uint(value, c->nchannels);
+        break;
+    case PROP_MUTE:
+        g_value_set_boolean(value, c->mute);
+        break;
+    default:
+        G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
+        break;
+    }
+}
+
+static void spice_record_channel_set_property(GObject      *gobject,
+                                              guint         prop_id,
+                                              const GValue *value,
+                                              GParamSpec   *pspec)
+{
+    switch (prop_id) {
+    case PROP_VOLUME:
+        /* TODO: request guest volume change */
+        break;
+    case PROP_MUTE:
+        /* TODO: request guest mute change */
+        break;
+    default:
+        G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
+        break;
+    }
+}
+
 static void spice_record_channel_class_init(SpiceRecordChannelClass *klass)
 {
     GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
     SpiceChannelClass *channel_class = SPICE_CHANNEL_CLASS(klass);
 
     gobject_class->finalize     = spice_record_channel_finalize;
+    gobject_class->get_property = spice_record_channel_get_property;
+    gobject_class->set_property = spice_record_channel_set_property;
     channel_class->handle_msg   = spice_record_handle_msg;
     channel_class->channel_up   = channel_up;
 
+    g_object_class_install_property
+        (gobject_class, PROP_NCHANNELS,
+         g_param_spec_uint("nchannels",
+                           "Number of Channels",
+                           "Number of Channels",
+                           0, G_MAXUINT8, 2,
+                           G_PARAM_READWRITE |
+                           G_PARAM_STATIC_STRINGS));
+
+    g_object_class_install_property
+        (gobject_class, PROP_VOLUME,
+         g_param_spec_pointer("volume",
+                              "Playback volume",
+                              "",
+                              G_PARAM_READWRITE |
+                              G_PARAM_STATIC_STRINGS));
+
+    g_object_class_install_property
+        (gobject_class, PROP_MUTE,
+         g_param_spec_boolean("mute",
+                              "Mute",
+                              "Mute",
+                              FALSE,
+                              G_PARAM_READWRITE |
+                              G_PARAM_STATIC_STRINGS));
     /**
      * SpiceRecordChannel::record-start:
      * @channel: the #SpiceRecordChannel that emitted the signal
@@ -153,7 +238,7 @@ static void spice_record_channel_class_init(SpiceRecordChannelClass *klass)
                      G_TYPE_NONE,
                      0);
 
-    g_type_class_add_private(klass, sizeof(spice_record_channel));
+    g_type_class_add_private(klass, sizeof(SpiceRecordChannelPrivate));
 }
 
 /* signal trampoline---------------------------------------------------------- */
@@ -190,14 +275,10 @@ static void do_emit_main_context(GObject *object, int signum, gpointer params)
 static void spice_record_mode(SpiceRecordChannel *channel, uint32_t time,
                               uint32_t mode, uint8_t *data, uint32_t data_size)
 {
-    spice_record_channel *rc;
-    spice_channel *c;
     SpiceMsgcRecordMode m = {0, };
-    spice_msg_out *msg;
+    SpiceMsgOut *msg;
 
     g_return_if_fail(channel != NULL);
-    rc = channel->priv;
-    c = SPICE_CHANNEL(channel)->priv;
 
     m.mode = mode;
     m.time = time;
@@ -213,7 +294,7 @@ static void spice_record_mode(SpiceRecordChannel *channel, uint32_t time,
 /* coroutine context */
 static void channel_up(SpiceChannel *channel)
 {
-    spice_record_channel *rc;
+    SpiceRecordChannelPrivate *rc;
 
     rc = SPICE_RECORD_CHANNEL(channel)->priv;
     if (spice_channel_test_capability(channel, SPICE_RECORD_CAP_CELT_0_5_1)) {
@@ -226,14 +307,10 @@ static void channel_up(SpiceChannel *channel)
 /* main context */
 static void spice_record_start_mark(SpiceRecordChannel *channel, uint32_t time)
 {
-    spice_record_channel *rc;
-    spice_channel *c;
     SpiceMsgcRecordStartMark m = {0, };
-    spice_msg_out *msg;
+    SpiceMsgOut *msg;
 
     g_return_if_fail(channel != NULL);
-    rc = channel->priv;
-    c = SPICE_CHANNEL(channel)->priv;
 
     m.time = time;
 
@@ -255,8 +332,7 @@ static void spice_record_start_mark(SpiceRecordChannel *channel, uint32_t time)
 void spice_record_send_data(SpiceRecordChannel *channel, gpointer data,
                             gsize bytes, uint32_t time)
 {
-    spice_record_channel *rc;
-    spice_channel *c;
+    SpiceRecordChannelPrivate *rc;
     SpiceMsgcRecordPacket p = {0, };
     int celt_compressed_frame_bytes = FRAME_SIZE * CELT_BIT_RATE / 44100 / 8;
     uint8_t *celt_buf = NULL;
@@ -264,7 +340,6 @@ void spice_record_send_data(SpiceRecordChannel *channel, gpointer data,
     g_return_if_fail(channel != NULL);
 
     rc = channel->priv;
-    c = SPICE_CHANNEL(channel)->priv;
 
     if (!rc->started) {
         spice_record_mode(channel, time, rc->mode, NULL, 0);
@@ -279,7 +354,7 @@ void spice_record_send_data(SpiceRecordChannel *channel, gpointer data,
 
     while (bytes > 0) {
         gsize n, frame_size;
-        spice_msg_out *msg;
+        SpiceMsgOut *msg;
         uint8_t *frame;
 
         if (rc->last_frame_current > 0) {
@@ -333,9 +408,9 @@ void spice_record_send_data(SpiceRecordChannel *channel, gpointer data,
 /* ------------------------------------------------------------------ */
 
 /* coroutine context */
-static void record_handle_start(SpiceChannel *channel, spice_msg_in *in)
+static void record_handle_start(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    spice_record_channel *c = SPICE_RECORD_CHANNEL(channel)->priv;
+    SpiceRecordChannelPrivate *c = SPICE_RECORD_CHANNEL(channel)->priv;
     SpiceMsgRecordStart *start = spice_msg_in_parsed(in);
 
     SPICE_DEBUG("%s: fmt %d channels %d freq %d", __FUNCTION__,
@@ -380,31 +455,59 @@ static void record_handle_start(SpiceChannel *channel, spice_msg_in *in)
 }
 
 /* coroutine context */
-static void record_handle_stop(SpiceChannel *channel, spice_msg_in *in)
+static void record_handle_stop(SpiceChannel *channel, SpiceMsgIn *in)
 {
-    spice_record_channel *rc = SPICE_RECORD_CHANNEL(channel)->priv;
+    SpiceRecordChannelPrivate *rc = SPICE_RECORD_CHANNEL(channel)->priv;
 
     emit_main_context(channel, SPICE_RECORD_STOP);
     rc->started = FALSE;
 }
 
-static spice_msg_handler record_handlers[] = {
-    [ SPICE_MSG_SET_ACK ]                  = spice_channel_handle_set_ack,
-    [ SPICE_MSG_PING ]                     = spice_channel_handle_ping,
-    [ SPICE_MSG_NOTIFY ]                   = spice_channel_handle_notify,
-    [ SPICE_MSG_DISCONNECTING ]            = spice_channel_handle_disconnect,
-    [ SPICE_MSG_WAIT_FOR_CHANNELS ]        = spice_channel_handle_wait_for_channels,
-    [ SPICE_MSG_MIGRATE ]                  = spice_channel_handle_migrate,
+/* coroutine context */
+static void record_handle_set_volume(SpiceChannel *channel, SpiceMsgIn *in)
+{
+    SpiceRecordChannelPrivate *c = SPICE_RECORD_CHANNEL(channel)->priv;
+    SpiceMsgAudioVolume *vol = spice_msg_in_parsed(in);
 
+    g_free(c->volume);
+    c->nchannels = vol->nchannels;
+    c->volume = g_new(guint16, c->nchannels);
+    memcpy(c->volume, vol->volume, sizeof(guint16) * c->nchannels);
+    g_object_notify_main_context(G_OBJECT(channel), "volume");
+}
+
+/* coroutine context */
+static void record_handle_set_mute(SpiceChannel *channel, SpiceMsgIn *in)
+{
+    SpiceRecordChannelPrivate *c = SPICE_RECORD_CHANNEL(channel)->priv;
+    SpiceMsgAudioMute *m = spice_msg_in_parsed(in);
+
+    c->mute = m->mute;
+    g_object_notify_main_context(G_OBJECT(channel), "mute");
+}
+
+static const spice_msg_handler record_handlers[] = {
     [ SPICE_MSG_RECORD_START ]             = record_handle_start,
     [ SPICE_MSG_RECORD_STOP ]              = record_handle_stop,
+    [ SPICE_MSG_RECORD_VOLUME ]            = record_handle_set_volume,
+    [ SPICE_MSG_RECORD_MUTE ]              = record_handle_set_mute,
 };
 
 /* coroutine context */
-static void spice_record_handle_msg(SpiceChannel *channel, spice_msg_in *msg)
+static void spice_record_handle_msg(SpiceChannel *channel, SpiceMsgIn *msg)
 {
     int type = spice_msg_in_type(msg);
+
+    SpiceChannelClass *parent_class;
+
     g_return_if_fail(type < SPICE_N_ELEMENTS(record_handlers));
-    g_return_if_fail(record_handlers[type] != NULL);
-    record_handlers[type](channel, msg);
+
+    parent_class = SPICE_CHANNEL_CLASS(spice_record_channel_parent_class);
+
+    if (record_handlers[type] != NULL)
+        record_handlers[type](channel, msg);
+    else if (parent_class->handle_msg)
+        parent_class->handle_msg(channel, msg);
+    else
+        g_return_if_reached();
 }
