@@ -30,6 +30,7 @@
 #include "spice-types.h"
 #include "spice-session.h"
 #include "spice-channel.h"
+#include "spice-option.h"
 
 #include "channel-main.h"
 #include "channel-display.h"
@@ -37,5 +38,21 @@
 #include "channel-inputs.h"
 #include "channel-playback.h"
 #include "channel-record.h"
+#include "channel-smartcard.h"
+
+#define SPICE_CLIENT_ERROR spice_client_error_quark()
+
+/**
+ * SpiceClientError:
+ * @SPICE_CLIENT_ERROR_FAILED: generic error code
+ *
+ * Error codes returned by spice-client API.
+ */
+typedef enum
+{
+    SPICE_CLIENT_ERROR_FAILED,
+} SpiceClientError;
+
+GQuark spice_client_error_quark(void);
 
 #endif /* __SPICE_CLIENT_CLIENT_H__ */
