@@ -372,7 +372,9 @@ g_cclosure_user_marshal_BOOLEAN__UINT (GClosure     *closure,
   g_value_set_boolean (return_value, v_return);
 }
 
-/* VOID:UINT,POINTER,UINT (spice-marshal.txt:9) */
+/* VOID:BOXED (spice-marshal.txt:9) */
+
+/* VOID:UINT,POINTER,UINT (spice-marshal.txt:10) */
 void
 g_cclosure_user_marshal_VOID__UINT_POINTER_UINT (GClosure     *closure,
                                                  GValue       *return_value G_GNUC_UNUSED,
@@ -411,3 +413,128 @@ g_cclosure_user_marshal_VOID__UINT_POINTER_UINT (GClosure     *closure,
             data2);
 }
 
+/* VOID:UINT,UINT,POINTER,UINT (spice-marshal.txt:11) */
+void
+g_cclosure_user_marshal_VOID__UINT_UINT_POINTER_UINT (GClosure     *closure,
+                                                      GValue       *return_value G_GNUC_UNUSED,
+                                                      guint         n_param_values,
+                                                      const GValue *param_values,
+                                                      gpointer      invocation_hint G_GNUC_UNUSED,
+                                                      gpointer      marshal_data)
+{
+  typedef void (*GMarshalFunc_VOID__UINT_UINT_POINTER_UINT) (gpointer     data1,
+                                                             guint        arg_1,
+                                                             guint        arg_2,
+                                                             gpointer     arg_3,
+                                                             guint        arg_4,
+                                                             gpointer     data2);
+  register GMarshalFunc_VOID__UINT_UINT_POINTER_UINT callback;
+  register GCClosure *cc = (GCClosure*) closure;
+  register gpointer data1, data2;
+
+  g_return_if_fail (n_param_values == 5);
+
+  if (G_CCLOSURE_SWAP_DATA (closure))
+    {
+      data1 = closure->data;
+      data2 = g_value_peek_pointer (param_values + 0);
+    }
+  else
+    {
+      data1 = g_value_peek_pointer (param_values + 0);
+      data2 = closure->data;
+    }
+  callback = (GMarshalFunc_VOID__UINT_UINT_POINTER_UINT) (marshal_data ? marshal_data : cc->callback);
+
+  callback (data1,
+            g_marshal_value_peek_uint (param_values + 1),
+            g_marshal_value_peek_uint (param_values + 2),
+            g_marshal_value_peek_pointer (param_values + 3),
+            g_marshal_value_peek_uint (param_values + 4),
+            data2);
+}
+
+/* BOOLEAN:UINT,POINTER,UINT (spice-marshal.txt:12) */
+void
+g_cclosure_user_marshal_BOOLEAN__UINT_POINTER_UINT (GClosure     *closure,
+                                                    GValue       *return_value G_GNUC_UNUSED,
+                                                    guint         n_param_values,
+                                                    const GValue *param_values,
+                                                    gpointer      invocation_hint G_GNUC_UNUSED,
+                                                    gpointer      marshal_data)
+{
+  typedef gboolean (*GMarshalFunc_BOOLEAN__UINT_POINTER_UINT) (gpointer     data1,
+                                                               guint        arg_1,
+                                                               gpointer     arg_2,
+                                                               guint        arg_3,
+                                                               gpointer     data2);
+  register GMarshalFunc_BOOLEAN__UINT_POINTER_UINT callback;
+  register GCClosure *cc = (GCClosure*) closure;
+  register gpointer data1, data2;
+  gboolean v_return;
+
+  g_return_if_fail (return_value != NULL);
+  g_return_if_fail (n_param_values == 4);
+
+  if (G_CCLOSURE_SWAP_DATA (closure))
+    {
+      data1 = closure->data;
+      data2 = g_value_peek_pointer (param_values + 0);
+    }
+  else
+    {
+      data1 = g_value_peek_pointer (param_values + 0);
+      data2 = closure->data;
+    }
+  callback = (GMarshalFunc_BOOLEAN__UINT_POINTER_UINT) (marshal_data ? marshal_data : cc->callback);
+
+  v_return = callback (data1,
+                       g_marshal_value_peek_uint (param_values + 1),
+                       g_marshal_value_peek_pointer (param_values + 2),
+                       g_marshal_value_peek_uint (param_values + 3),
+                       data2);
+
+  g_value_set_boolean (return_value, v_return);
+}
+
+/* BOOLEAN:UINT,UINT (spice-marshal.txt:13) */
+void
+g_cclosure_user_marshal_BOOLEAN__UINT_UINT (GClosure     *closure,
+                                            GValue       *return_value G_GNUC_UNUSED,
+                                            guint         n_param_values,
+                                            const GValue *param_values,
+                                            gpointer      invocation_hint G_GNUC_UNUSED,
+                                            gpointer      marshal_data)
+{
+  typedef gboolean (*GMarshalFunc_BOOLEAN__UINT_UINT) (gpointer     data1,
+                                                       guint        arg_1,
+                                                       guint        arg_2,
+                                                       gpointer     data2);
+  register GMarshalFunc_BOOLEAN__UINT_UINT callback;
+  register GCClosure *cc = (GCClosure*) closure;
+  register gpointer data1, data2;
+  gboolean v_return;
+
+  g_return_if_fail (return_value != NULL);
+  g_return_if_fail (n_param_values == 3);
+
+  if (G_CCLOSURE_SWAP_DATA (closure))
+    {
+      data1 = closure->data;
+      data2 = g_value_peek_pointer (param_values + 0);
+    }
+  else
+    {
+      data1 = g_value_peek_pointer (param_values + 0);
+      data2 = closure->data;
+    }
+  callback = (GMarshalFunc_BOOLEAN__UINT_UINT) (marshal_data ? marshal_data : cc->callback);
+
+  v_return = callback (data1,
+                       g_marshal_value_peek_uint (param_values + 1),
+                       g_marshal_value_peek_uint (param_values + 2),
+                       data2);
+
+  g_value_set_boolean (return_value, v_return);
+}
+
