@@ -20,6 +20,7 @@
 #include "spice-client.h"
 #include "spice-common.h"
 #include "spice-channel-priv.h"
+#include "spice-util-priv.h"
 
 #include "spice-session-priv.h"
 
@@ -33,19 +34,37 @@ struct channel {
     RingItem          link;
 };
 
-struct spice_session {
+struct _SpiceSessionPrivate {
     char              *host;
     char              *port;
     char              *tls_port;
     char              *password;
     char              *ca_file;
+    char              *ciphers;
     GByteArray        *pubkey;
     char              *cert_subject;
     guint             verify;
 
+    /* whether to enable smartcard event forwarding to the server */
+    gboolean          smartcard;
+
+    /* list of certificates to use for the software smartcard reader if
+     * enabled. For now, it has to contain exactly 3 certificates for
+     * the software reader to be functional
+     */
+    GStrv             smartcard_certificates;
+
+    /* path to the local certificate database to use to lookup the
+     * certificates stored in 'certificates'. If NULL, libcacard will
+     * fallback to using a default database.
+     */
+    char *            smartcard_db;
+    GStrv             disable_effects;
+    gint              color_depth;
+
     int               connection_id;
     int               protocol;
-    SpiceChannel      *cmain;
+    SpiceChannel      *cmain; /* weak reference */
     Ring              channels;
     guint32           mm_time;
     gboolean          client_provided_sockets;
@@ -54,6 +73,10 @@ struct spice_session {
     GList             *migration_left;
     SpiceSessionMigration migration_state;
     gboolean          disconnecting;
+
+    display_cache     images;
+    display_cache     palettes;
+    SpiceGlzDecoderWindow *glz_window;
 };
 
 /**
@@ -91,7 +114,7 @@ struct spice_session {
 /* gobject glue                                                       */
 
 #define SPICE_SESSION_GET_PRIVATE(obj) \
-    (G_TYPE_INSTANCE_GET_PRIVATE ((obj), SPICE_TYPE_SESSION, spice_session))
+    (G_TYPE_INSTANCE_GET_PRIVATE ((obj), SPICE_TYPE_SESSION, SpiceSessionPrivate))
 
 G_DEFINE_TYPE (SpiceSession, spice_session, G_TYPE_OBJECT);
 
@@ -103,6 +126,7 @@ enum {
     PROP_TLS_PORT,
     PROP_PASSWORD,
     PROP_CA_FILE,
+    PROP_CIPHERS,
     PROP_IPV4,
     PROP_IPV6,
     PROP_PROTOCOL,
@@ -112,6 +136,11 @@ enum {
     PROP_CERT_SUBJECT,
     PROP_VERIFY,
     PROP_MIGRATION_STATE,
+    PROP_SMARTCARD,
+    PROP_SMARTCARD_CERTIFICATES,
+    PROP_SMARTCARD_DB,
+    PROP_DISABLE_EFFECTS,
+    PROP_COLOR_DEPTH,
 };
 
 /* signals */
@@ -123,21 +152,26 @@ enum {
 
 static guint signals[SPICE_SESSION_LAST_SIGNAL];
 
+
 static void spice_session_init(SpiceSession *session)
 {
-    spice_session *s;
+    SpiceSessionPrivate *s;
 
+    SPICE_DEBUG("New session (compiled from package " PACKAGE_STRING ")");
     s = session->priv = SPICE_SESSION_GET_PRIVATE(session);
     memset(s, 0, sizeof(*s));
 
     ring_init(&s->channels);
+    cache_init(&s->images, "image");
+    cache_init(&s->palettes, "palette");
+    s->glz_window = glz_decoder_window_new();
 }
 
 static void
 spice_session_dispose(GObject *gobject)
 {
     SpiceSession *session = SPICE_SESSION(gobject);
-    spice_session *s = session->priv;
+    SpiceSessionPrivate *s = session->priv;
 
     SPICE_DEBUG("session dispose");
 
@@ -159,11 +193,40 @@ spice_session_dispose(GObject *gobject)
         G_OBJECT_CLASS(spice_session_parent_class)->dispose(gobject);
 }
 
+G_GNUC_INTERNAL
+void spice_session_palettes_clear(SpiceSession *session)
+{
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
+    g_return_if_fail(s != NULL);
+
+    for (;;) {
+        display_cache_item *item = cache_get_lru(&s->palettes);
+        if (item == NULL)
+            break;
+        cache_del(&s->palettes, item);
+    }
+}
+
+G_GNUC_INTERNAL
+void spice_session_images_clear(SpiceSession *session)
+{
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
+    g_return_if_fail(s != NULL);
+
+    for (;;) {
+        display_cache_item *item = cache_get_lru(&s->images);
+        if (item == NULL)
+            break;
+        pixman_image_unref(item->ptr);
+        cache_del(&s->images, item);
+    }
+}
+
 static void
 spice_session_finalize(GObject *gobject)
 {
     SpiceSession *session = SPICE_SESSION(gobject);
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
 
     /* release stuff */
     g_free(s->host);
@@ -171,7 +234,15 @@ spice_session_finalize(GObject *gobject)
     g_free(s->tls_port);
     g_free(s->password);
     g_free(s->ca_file);
+    g_free(s->ciphers);
     g_free(s->cert_subject);
+    g_strfreev(s->smartcard_certificates);
+    g_free(s->smartcard_db);
+    g_strfreev(s->disable_effects);
+
+    spice_session_palettes_clear(session);
+    spice_session_images_clear(session);
+    glz_decoder_window_destroy(s->glz_window);
 
     if (s->pubkey)
         g_byte_array_unref(s->pubkey);
@@ -183,7 +254,7 @@ spice_session_finalize(GObject *gobject)
 
 static int spice_uri_create(SpiceSession *session, char *dest, int len)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
     int pos = 0;
 
     if (s->host == NULL || (s->port == NULL && s->tls_port == NULL)) {
@@ -200,7 +271,7 @@ static int spice_uri_create(SpiceSession *session, char *dest, int len)
 
 static int spice_uri_parse(SpiceSession *session, const char *original_uri)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
     char host[128], key[32], value[128];
     char *port = NULL, *tls_port = NULL, *uri = NULL, *password = NULL;
     int len, pos = 0;
@@ -268,7 +339,7 @@ static void spice_session_get_property(GObject    *gobject,
                                        GParamSpec *pspec)
 {
     SpiceSession *session = SPICE_SESSION(gobject);
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
     char buf[256];
     int len;
 
@@ -288,6 +359,9 @@ static void spice_session_get_property(GObject    *gobject,
     case PROP_CA_FILE:
         g_value_set_string(value, s->ca_file);
 	break;
+    case PROP_CIPHERS:
+        g_value_set_string(value, s->ciphers);
+	break;
     case PROP_PROTOCOL:
         g_value_set_int(value, s->protocol);
 	break;
@@ -310,6 +384,21 @@ static void spice_session_get_property(GObject    *gobject,
     case PROP_MIGRATION_STATE:
         g_value_set_enum(value, s->migration_state);
         break;
+    case PROP_SMARTCARD:
+        g_value_set_boolean(value, s->smartcard);
+        break;
+    case PROP_SMARTCARD_CERTIFICATES:
+        g_value_set_boxed(value, s->smartcard_certificates);
+        break;
+    case PROP_SMARTCARD_DB:
+        g_value_set_string(value, s->smartcard_db);
+        break;
+    case PROP_DISABLE_EFFECTS:
+        g_value_set_boxed(value, s->disable_effects);
+        break;
+    case PROP_COLOR_DEPTH:
+        g_value_set_int(value, s->color_depth);
+        break;
     default:
 	G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
 	break;
@@ -322,7 +411,7 @@ static void spice_session_set_property(GObject      *gobject,
                                        GParamSpec   *pspec)
 {
     SpiceSession *session = SPICE_SESSION(gobject);
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
     const char *str;
 
     switch (prop_id) {
@@ -346,6 +435,10 @@ static void spice_session_set_property(GObject      *gobject,
         g_free(s->ca_file);
         s->ca_file = g_value_dup_string(value);
         break;
+    case PROP_CIPHERS:
+        g_free(s->ciphers);
+        s->ciphers = g_value_dup_string(value);
+        break;
     case PROP_PROTOCOL:
         s->protocol = g_value_get_int(value);
         break;
@@ -375,6 +468,24 @@ static void spice_session_set_property(GObject      *gobject,
     case PROP_MIGRATION_STATE:
         s->migration_state = g_value_get_enum(value);
         break;
+    case PROP_SMARTCARD:
+        s->smartcard = g_value_get_boolean(value);
+        break;
+    case PROP_SMARTCARD_CERTIFICATES:
+        g_strfreev(s->smartcard_certificates);
+        s->smartcard_certificates = g_value_dup_boxed(value);
+        break;
+    case PROP_SMARTCARD_DB:
+        g_free(s->smartcard_db);
+        s->smartcard_db = g_value_dup_string(value);
+        break;
+    case PROP_DISABLE_EFFECTS:
+        g_strfreev(s->disable_effects);
+        s->disable_effects = g_value_dup_boxed(value);
+        break;
+    case PROP_COLOR_DEPTH:
+        s->color_depth = g_value_get_int(value);
+        break;
     default:
         G_OBJECT_WARN_INVALID_PROPERTY_ID(gobject, prop_id, pspec);
         break;
@@ -390,6 +501,12 @@ static void spice_session_class_init(SpiceSessionClass *klass)
     gobject_class->get_property = spice_session_get_property;
     gobject_class->set_property = spice_session_set_property;
 
+    /**
+     * SpiceSession:host:
+     *
+     * URL of the SPICE host to connect to
+     *
+     **/
     g_object_class_install_property
         (gobject_class, PROP_HOST,
          g_param_spec_string("host",
@@ -400,6 +517,12 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                              G_PARAM_CONSTRUCT |
                              G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceSession:port:
+     *
+     * Port to connect to for unencrypted sessions
+     *
+     **/
     g_object_class_install_property
         (gobject_class, PROP_PORT,
          g_param_spec_string("port",
@@ -409,6 +532,12 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                              G_PARAM_READWRITE |
                              G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceSession:tls-port:
+     *
+     * Port to connect to for TLS sessions
+     *
+     **/
     g_object_class_install_property
         (gobject_class, PROP_TLS_PORT,
          g_param_spec_string("tls-port",
@@ -418,6 +547,12 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                              G_PARAM_READWRITE |
                              G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceSession:password:
+     *
+     * TLS password to use
+     *
+     **/
     g_object_class_install_property
         (gobject_class, PROP_PASSWORD,
          g_param_spec_string("password",
@@ -427,6 +562,13 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                              G_PARAM_READWRITE |
                              G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceSession:ca-file:
+     *
+     * File holding the CA certificates for the host the client is
+     * connecting to
+     *
+     **/
     g_object_class_install_property
         (gobject_class, PROP_CA_FILE,
          g_param_spec_string("ca-file",
@@ -436,6 +578,25 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                              G_PARAM_READWRITE |
                              G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceSession:ciphers:
+     *
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_CIPHERS,
+         g_param_spec_string("ciphers",
+                             "Ciphers",
+                             "SSL cipher list",
+                             NULL,
+                             G_PARAM_READWRITE |
+                             G_PARAM_STATIC_STRINGS));
+
+    /**
+     * SpiceSession:protocol:
+     *
+     * Version of the SPICE protocol to use
+     *
+     **/
     g_object_class_install_property
         (gobject_class, PROP_PROTOCOL,
          g_param_spec_int("protocol",
@@ -446,6 +607,13 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                           G_PARAM_CONSTRUCT |
                           G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceSession:uri:
+     *
+     * URI of the SPICE host to connect to. The URI is of the form
+     * spice://hostname?port=XXX or spice://hostname?tls_port=XXX
+     *
+     **/
     g_object_class_install_property
         (gobject_class, PROP_URI,
          g_param_spec_string("uri",
@@ -455,6 +623,10 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                              G_PARAM_READWRITE |
                              G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceSession:client-sockets:
+     *
+     **/
     g_object_class_install_property
         (gobject_class, PROP_CLIENT_SOCKETS,
          g_param_spec_boolean("client-sockets",
@@ -464,6 +636,10 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                           G_PARAM_READWRITE |
                           G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceSession:pubkey:
+     *
+     **/
     g_object_class_install_property
         (gobject_class, PROP_PUBKEY,
          g_param_spec_boxed("pubkey",
@@ -473,6 +649,10 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                             G_PARAM_READWRITE |
                             G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceSession:cert-subject:
+     *
+     **/
     g_object_class_install_property
         (gobject_class, PROP_CERT_SUBJECT,
          g_param_spec_string("cert-subject",
@@ -482,6 +662,12 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                              G_PARAM_READWRITE |
                              G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceSession:verify:
+     *
+     * #SpiceSessionVerify bit field indicating which parts of the peer
+     * certificate should be checked
+     **/
     g_object_class_install_property
         (gobject_class, PROP_VERIFY,
          g_param_spec_flags("verify",
@@ -493,6 +679,13 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                             G_PARAM_CONSTRUCT |
                             G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceSession:migration-state:
+     *
+     * #SpiceSessionMigration bit field indicating if a migration is in
+     * progress
+     *
+     **/
     g_object_class_install_property
         (gobject_class, PROP_MIGRATION_STATE,
          g_param_spec_enum("migration-state",
@@ -503,6 +696,97 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                            G_PARAM_READABLE |
                            G_PARAM_STATIC_STRINGS));
 
+    /**
+     * SpiceSession:disable-effects:
+     *
+     * A comma-separated list of effects to disable. The settings will
+     * be applied on new display channels. The following effets can be
+     * disabled "wallpaper", "font-smooth", "animation", and "all",
+     * which will disable all the effects. If NULL, don't apply changes.
+     *
+     * Since: 0.7
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_DISABLE_EFFECTS,
+         g_param_spec_boxed ("disable-effects",
+                             "Disable effects",
+                             "Comma-separated effects to disable",
+                             G_TYPE_STRV,
+                             G_PARAM_READWRITE |
+                             G_PARAM_STATIC_STRINGS));
+
+    /**
+     * SpiceSession:color-depth:
+     *
+     * Display color depth to set on new display channels. If 0, don't set.
+     *
+     * Since: 0.7
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_COLOR_DEPTH,
+         g_param_spec_int("color-depth",
+                          "Color depth",
+                          "Display channel color depth",
+                          0, 32, 0,
+                          G_PARAM_READWRITE |
+                          G_PARAM_STATIC_STRINGS));
+
+    /**
+     * SpiceSession:enable-smartcard:
+     *
+     * If set to TRUE, the smartcard channel will be enabled and smartcard
+     * events will be forwarded to the guest
+     *
+     * Since: 0.7
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_SMARTCARD,
+         g_param_spec_boolean("enable-smartcard",
+                          "Enable smartcard event forwarding",
+                          "Forward smartcard events to the SPICE server",
+                          FALSE,
+                          G_PARAM_READWRITE |
+                          G_PARAM_STATIC_STRINGS));
+
+    /**
+     * SpiceSession:smartcard-certificates:
+     *
+     * This property is used when one wants to simulate a smartcard with no
+     * hardware smartcard reader. If it's set to a NULL-terminated string
+     * array containing the names of 3 valid certificates, these will be
+     * used to simulate a smartcard in the guest
+     * @see_also: spice_smartcard_manager_insert_card()
+     *
+     * Since: 0.7
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_SMARTCARD_CERTIFICATES,
+         g_param_spec_boxed("smartcard-certificates",
+                            "Smartcard certificates",
+                            "Smartcard certificates for software-based smartcards",
+                            G_TYPE_STRV,
+                            G_PARAM_READABLE |
+                            G_PARAM_WRITABLE |
+                            G_PARAM_STATIC_STRINGS));
+
+    /**
+     * SpiceSession:smartcard-db:
+     *
+     * Path to the NSS certificate database containing the certificates to
+     * use to simulate a software smartcard
+     *
+     * Since: 0.7
+     **/
+    g_object_class_install_property
+        (gobject_class, PROP_SMARTCARD_DB,
+         g_param_spec_string("smartcard-db",
+                              "Smartcard certificate database",
+                              "Path to the database for smartcard certificates",
+                              NULL,
+                              G_PARAM_READABLE |
+                              G_PARAM_WRITABLE |
+                              G_PARAM_STATIC_STRINGS));
+
     /**
      * SpiceSession::channel-new:
      * @session: the session that emitted the signal
@@ -539,7 +823,7 @@ static void spice_session_class_init(SpiceSessionClass *klass)
                      1,
                      SPICE_TYPE_CHANNEL);
 
-    g_type_class_add_private(klass, sizeof(spice_session));
+    g_type_class_add_private(klass, sizeof(SpiceSessionPrivate));
 }
 
 /* ------------------------------------------------------------------ */
@@ -564,16 +848,27 @@ SpiceSession *spice_session_new_from_session(SpiceSession *session)
                                                     "host", NULL,
                                                     "ca-file", NULL,
                                                     NULL));
-    spice_session *c = copy->priv, *s = session->priv;
+    SpiceSessionPrivate *c = copy->priv, *s = session->priv;
+
+    g_warn_if_fail (c->host == NULL);
+    g_warn_if_fail (c->tls_port == NULL);
+    g_warn_if_fail (c->password == NULL);
+    g_warn_if_fail (c->ca_file == NULL);
+    g_warn_if_fail (c->ciphers == NULL);
+    g_warn_if_fail (c->cert_subject == NULL);
+    g_warn_if_fail (c->pubkey == NULL);
 
     g_object_get(session,
                  "host", &c->host,
                  "tls-port", &c->tls_port,
                  "password", &c->password,
                  "ca-file", &c->ca_file,
+                 "ciphers", &c->ciphers,
                  "cert-subject", &c->cert_subject,
                  "pubkey", &c->pubkey,
                  "verify", &c->verify,
+                 "smartcard-certificates", &c->smartcard_certificates,
+                 "smartcard-db", &c->smartcard_db,
                  NULL);
 
     c->client_provided_sockets = s->client_provided_sockets;
@@ -594,15 +889,19 @@ SpiceSession *spice_session_new_from_session(SpiceSession *session)
  **/
 gboolean spice_session_connect(SpiceSession *session)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
 
     g_return_val_if_fail(s != NULL, FALSE);
 
     spice_session_disconnect(session);
+    s->disconnecting = FALSE;
 
     s->client_provided_sockets = FALSE;
+
+    g_warn_if_fail(s->cmain == NULL);
     s->cmain = spice_channel_new(session, SPICE_CHANNEL_MAIN, 0);
 
+    glz_decoder_window_clear(s->glz_window);
     return spice_channel_connect(s->cmain);
 }
 
@@ -619,7 +918,7 @@ gboolean spice_session_connect(SpiceSession *session)
  **/
 gboolean spice_session_open_fd(SpiceSession *session, int fd)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
 
     g_return_val_if_fail(s != NULL, FALSE);
     g_return_val_if_fail(fd >= 0, FALSE);
@@ -627,6 +926,8 @@ gboolean spice_session_open_fd(SpiceSession *session, int fd)
     spice_session_disconnect(session);
 
     s->client_provided_sockets = TRUE;
+
+    g_warn_if_fail(s->cmain == NULL);
     s->cmain = spice_channel_new(session, SPICE_CHANNEL_MAIN, 0);
     return spice_channel_open_fd(s->cmain, fd);
 }
@@ -634,7 +935,7 @@ gboolean spice_session_open_fd(SpiceSession *session, int fd)
 G_GNUC_INTERNAL
 gboolean spice_session_get_client_provided_socket(SpiceSession *session)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
 
     g_return_val_if_fail(s != NULL, FALSE);
     return s->client_provided_sockets;
@@ -643,7 +944,7 @@ gboolean spice_session_get_client_provided_socket(SpiceSession *session)
 G_GNUC_INTERNAL
 void spice_session_switching_disconnect(SpiceSession *session)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
     struct channel *item;
     RingItem *ring, *next;
 
@@ -665,8 +966,8 @@ void spice_session_switching_disconnect(SpiceSession *session)
 G_GNUC_INTERNAL
 void spice_session_set_migration(SpiceSession *session, SpiceSession *migration)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
-    spice_session *m = SPICE_SESSION_GET_PRIVATE(migration);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *m = SPICE_SESSION_GET_PRIVATE(migration);
     gchar *tmp;
 
     g_return_if_fail(s != NULL);
@@ -700,7 +1001,7 @@ SpiceChannel* get_channel_by_id_and_type(SpiceSession *session,
                                          gint id, gint type)
 {
     RingItem *ring, *next;
-    spice_session *s = session->priv;
+    SpiceSessionPrivate *s = session->priv;
     struct channel *c;
 
     g_return_val_if_fail(s != NULL, NULL);
@@ -726,7 +1027,7 @@ SpiceChannel* get_channel_by_id_and_type(SpiceSession *session,
 G_GNUC_INTERNAL
 void spice_session_abort_migration(SpiceSession *session)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
     RingItem *ring, *next;
     struct channel *c;
 
@@ -759,7 +1060,7 @@ void spice_session_abort_migration(SpiceSession *session)
 G_GNUC_INTERNAL
 void spice_session_channel_migrate(SpiceSession *session, SpiceChannel *channel)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
     SpiceChannel *c;
     gint id, type;
 
@@ -794,7 +1095,7 @@ void spice_session_channel_migrate(SpiceSession *session, SpiceChannel *channel)
  **/
 void spice_session_disconnect(SpiceSession *session)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
     struct channel *item;
     RingItem *ring, *next;
 
@@ -814,20 +1115,23 @@ void spice_session_disconnect(SpiceSession *session)
     }
 
     s->connection_id = 0;
-    s->disconnecting = FALSE;
+    /* we leave disconnecting = TRUE, so that spice_channel_destroy()
+       is not called multiple times on channels that are in pending
+       destroy state. */
 }
 
 /**
  * spice_session_get_channels:
- * @session:
+ * @session: a #SpiceSession
  *
  * Get the list of current channels associated with this @session.
  *
- * Returns: a #GList of unowned SpiceChannels.
+ * Returns: (element-type SpiceChannel) (transfer container): a #GList
+ *          of unowned #SpiceChannel channels.
  **/
 GList *spice_session_get_channels(SpiceSession *session)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
     struct channel *item;
     GList *list = NULL;
     RingItem *ring;
@@ -884,7 +1188,7 @@ static GSocket *channel_connect_socket(GSocketAddress *sockaddr,
 G_GNUC_INTERNAL
 GSocket* spice_session_channel_open_host(SpiceSession *session, gboolean use_tls)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
     GSocketConnectable *addr;
     GSocketAddressEnumerator *enumerator;
     GSocketAddress *sockaddr;
@@ -924,7 +1228,7 @@ GSocket* spice_session_channel_open_host(SpiceSession *session, gboolean use_tls
 G_GNUC_INTERNAL
 void spice_session_channel_new(SpiceSession *session, SpiceChannel *channel)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
     struct channel *item;
 
     g_return_if_fail(s != NULL);
@@ -933,13 +1237,26 @@ void spice_session_channel_new(SpiceSession *session, SpiceChannel *channel)
     item = spice_new0(struct channel, 1);
     item->channel = channel;
     ring_add(&s->channels, &item->link);
+
+    if (SPICE_IS_MAIN_CHANNEL(channel)) {
+        gboolean all = spice_strv_contains(s->disable_effects, "all");
+
+        g_object_set(channel,
+                     "disable-wallpaper", all || spice_strv_contains(s->disable_effects, "wallpaper"),
+                     "disable-font-smooth", all || spice_strv_contains(s->disable_effects, "font-smooth"),
+                     "disable-animation", all || spice_strv_contains(s->disable_effects, "animation"),
+                     NULL);
+        if (s->color_depth != 0)
+            g_object_set(channel, "color-depth", s->color_depth, NULL);
+    }
+
     g_signal_emit(session, signals[SPICE_SESSION_CHANNEL_NEW], 0, channel);
 }
 
 G_GNUC_INTERNAL
 void spice_session_channel_destroy(SpiceSession *session, SpiceChannel *channel)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
     struct channel *item = NULL;
     RingItem *ring, *next;
 
@@ -953,6 +1270,10 @@ void spice_session_channel_destroy(SpiceSession *session, SpiceChannel *channel)
          ring = next) {
         next = ring_next(&s->channels, ring);
         item = SPICE_CONTAINEROF(ring, struct channel, link);
+        if (item->channel == s->cmain) {
+            SPICE_DEBUG("the session lost the main channel");
+            s->cmain = NULL;
+        }
         if (item->channel == channel) {
             ring_remove(&item->link);
             free(item);
@@ -967,7 +1288,7 @@ void spice_session_channel_destroy(SpiceSession *session, SpiceChannel *channel)
 G_GNUC_INTERNAL
 void spice_session_set_connection_id(SpiceSession *session, int id)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
 
     g_return_if_fail(s != NULL);
 
@@ -977,20 +1298,20 @@ void spice_session_set_connection_id(SpiceSession *session, int id)
 G_GNUC_INTERNAL
 int spice_session_get_connection_id(SpiceSession *session)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
 
     g_return_val_if_fail(s != NULL, -1);
 
     return s->connection_id;
 }
 
-#if !GLIB_CHECK_VERSION(2,28,0)
-static guint64 g_get_monotonic_clock(void)
+#if !GLIB_CHECK_VERSION(2,27,2)
+static guint64 g_get_monotonic_time(void)
 {
     GTimeVal tv;
 
     /* TODO: support real monotonic clock? */
-    g_get_current_time (&tv);
+    g_get_current_time(&tv);
 
     return (((gint64) tv.tv_sec) * 1000000) + tv.tv_usec;
 }
@@ -999,23 +1320,25 @@ static guint64 g_get_monotonic_clock(void)
 G_GNUC_INTERNAL
 guint32 spice_session_get_mm_time(SpiceSession *session)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
 
     g_return_val_if_fail(s != NULL, 0);
 
-    return s->mm_time + (g_get_monotonic_clock() - s->mm_time_at_clock) / 1000;
+    /* FIXME: we may want to estimate the drift of clocks, and well,
+       do something better than this trivial approach */
+    return s->mm_time + (g_get_monotonic_time() - s->mm_time_at_clock) / 1000;
 }
 
 G_GNUC_INTERNAL
 void spice_session_set_mm_time(SpiceSession *session, guint32 time)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
 
     g_return_if_fail(s != NULL);
-    SPICE_DEBUG("set mm time: %u", time);
 
     s->mm_time = time;
-    s->mm_time_at_clock = g_get_monotonic_clock();
+    s->mm_time_at_clock = g_get_monotonic_time();
+    SPICE_DEBUG("set mm time: %u", spice_session_get_mm_time(session));
 }
 
 G_GNUC_INTERNAL
@@ -1035,7 +1358,7 @@ void spice_session_set_port(SpiceSession *session, int port, gboolean tls)
 G_GNUC_INTERNAL
 void spice_session_get_pubkey(SpiceSession *session, guint8 **pubkey, guint *size)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
 
     g_return_if_fail(s != NULL);
     g_return_if_fail(pubkey != NULL);
@@ -1048,7 +1371,7 @@ void spice_session_get_pubkey(SpiceSession *session, guint8 **pubkey, guint *siz
 G_GNUC_INTERNAL
 guint spice_session_get_verify(SpiceSession *session)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
 
     g_return_val_if_fail(s != NULL, 0);
     return s->verify;
@@ -1057,9 +1380,72 @@ guint spice_session_get_verify(SpiceSession *session)
 G_GNUC_INTERNAL
 void spice_session_set_migration_state(SpiceSession *session, SpiceSessionMigration state)
 {
-    spice_session *s = SPICE_SESSION_GET_PRIVATE(session);
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
 
     g_return_if_fail(s != NULL);
     s->migration_state = state;
     g_object_notify(G_OBJECT(session), "migration-state");
 }
+
+G_GNUC_INTERNAL
+const gchar* spice_session_get_password(SpiceSession *session)
+{
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
+
+    g_return_val_if_fail(s != NULL, NULL);
+    return s->password;
+}
+
+G_GNUC_INTERNAL
+const gchar* spice_session_get_host(SpiceSession *session)
+{
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
+
+    g_return_val_if_fail(s != NULL, NULL);
+    return s->host;
+}
+
+G_GNUC_INTERNAL
+const gchar* spice_session_get_cert_subject(SpiceSession *session)
+{
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
+
+    g_return_val_if_fail(s != NULL, NULL);
+    return s->cert_subject;
+}
+
+G_GNUC_INTERNAL
+const gchar* spice_session_get_ciphers(SpiceSession *session)
+{
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
+
+    g_return_val_if_fail(s != NULL, NULL);
+    return s->ciphers;
+}
+
+G_GNUC_INTERNAL
+const gchar* spice_session_get_ca_file(SpiceSession *session)
+{
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
+
+    g_return_val_if_fail(s != NULL, NULL);
+    return s->ca_file;
+}
+
+G_GNUC_INTERNAL
+void spice_session_get_caches(SpiceSession *session,
+                              display_cache **images,
+                              display_cache **palettes,
+                              SpiceGlzDecoderWindow **glz_window)
+{
+    SpiceSessionPrivate *s = SPICE_SESSION_GET_PRIVATE(session);
+
+    g_return_if_fail(s != NULL);
+
+    if (images)
+        *images = &s->images;
+    if (palettes)
+        *palettes = &s->palettes;
+    if (glz_window)
+        *glz_window = s->glz_window;
+}
