@@ -30,14 +30,20 @@ G_BEGIN_DECLS
 #include <gdk/gdkx.h>
 #endif
 
+#ifdef WIN32
+#include <windows.h>
+#endif
+
 #include "spice-widget.h"
 #include "spice-common.h"
 #include <spice/vd_agent.h>
 
 #define SPICE_DISPLAY_GET_PRIVATE(obj)                                  \
-    (G_TYPE_INSTANCE_GET_PRIVATE((obj), SPICE_TYPE_DISPLAY, spice_display))
+    (G_TYPE_INSTANCE_GET_PRIVATE((obj), SPICE_TYPE_DISPLAY, SpiceDisplayPrivate))
 
-struct spice_display {
+#define CLIPBOARD_LAST (VD_AGENT_CLIPBOARD_SELECTION_SECONDARY + 1)
+
+struct _SpiceDisplayPrivate {
     gint                    channel_id;
 
     /* options */
@@ -52,7 +58,6 @@ struct spice_display {
     gint                    shmid;
     gpointer                data_origin; /* the original display image data */
     gpointer                data; /* converted if necessary to 32 bits */
-    gboolean                clipboard_by_guest; /* hack? */
 
     gint                    ww, wh, mx, my;
 
@@ -69,25 +74,29 @@ struct spice_display {
     GC                      gc;
 #else
     cairo_surface_t         *ximage;
-    cairo_surface_t         *ximage_cache;
 #endif
 
     GtkClipboard            *clipboard;
-    GtkTargetEntry          *clip_targets;
-    guint                   nclip_targets;
-    bool                    clip_hasdata;
-    bool                    clip_grabbed;
+    GtkClipboard            *clipboard_primary;
+    GtkTargetEntry          *clip_targets[CLIPBOARD_LAST];
+    guint                   nclip_targets[CLIPBOARD_LAST];
+    bool                    clip_hasdata[CLIPBOARD_LAST];
+    bool                    clip_grabbed[CLIPBOARD_LAST];
+    gboolean                clipboard_by_guest[CLIPBOARD_LAST];
+    gboolean                clipboard_selfgrab_pending[CLIPBOARD_LAST];
 
     SpiceSession            *session;
     SpiceMainChannel        *main;
     SpiceChannel            *display;
     SpiceCursorChannel      *cursor;
     SpiceInputsChannel      *inputs;
+    SpiceSmartcardChannel   *smartcard;
 
     enum SpiceMouseMode     mouse_mode;
     int                     mouse_grab_active;
     bool                    mouse_have_pointer;
     GdkCursor               *mouse_cursor;
+    GdkCursor               *show_cursor;
     int                     mouse_last_x;
     int                     mouse_last_y;
     int                     mouse_guest_x;
@@ -104,6 +113,9 @@ struct spice_display {
     SpiceGrabSequence         *grabseq; /* the configured key sequence */
     gboolean                *activeseq; /* the currently pressed keys */
     gint                    mark;
+#ifdef WIN32
+    HHOOK                   keyboard_hook;
+#endif
 };
 
 int      spicex_image_create                 (SpiceDisplay *display);
