@@ -251,14 +251,18 @@ def write_validate_pointer_item(writer, container, item, scope, parent_scope, st
 
             array_item = ItemInfo(target_type, "%s__array" % item.prefix, start)
             scope.variable_def("uint32_t", array_item.nw_size())
-            scope.variable_def("uint32_t", array_item.mem_size())
+            # don't create a variable that isn't used, fixes -Werror=unused-but-set-variable
+            need_mem_size = want_mem_size or (
+                want_extra_size and not item.member.has_attr("chunk")
+                and not target_type.is_cstring_length())
+            if need_mem_size:
+                scope.variable_def("uint32_t", array_item.mem_size())
             if target_type.is_cstring_length():
                 writer.assign(array_item.nw_size(), "spice_strnlen((char *)message_start + %s, message_end - (message_start + %s))" % (v, v))
                 writer.error_check("*(message_start + %s + %s) != 0" % (v, array_item.nw_size()))
-                writer.assign(array_item.mem_size(), array_item.nw_size())
             else:
                 write_validate_array_item(writer, container, array_item, scope, parent_scope, start,
-                                          True, True, False)
+                                          True, want_mem_size=need_mem_size, want_extra_size=False)
                 writer.error_check("message_start + %s + %s > message_end" % (v, array_item.nw_size()))
 
             if want_extra_size:
@@ -524,7 +528,7 @@ def write_validate_member(writer, container, member, parent_scope, start,
 def write_validate_container(writer, prefix, container, start, parent_scope, want_nw_size, want_mem_size, want_extra_size):
     for m in container.members:
         sub_want_nw_size = want_nw_size and not m.is_fixed_nw_size()
-        sub_want_mem_size = m.is_extra_size()
+        sub_want_mem_size = m.is_extra_size() and want_mem_size
         sub_want_extra_size = not m.is_extra_size() and m.contains_extra_size()
 
         defs = ["size_t"]
@@ -1007,14 +1011,21 @@ def write_msg_parser(writer, message):
     msg_type = message.c_type()
     msg_sizeof = message.sizeof()
 
+    want_mem_size = (len(message.members) != 1 or message.members[0].is_fixed_nw_size()
+                         or not message.members[0].is_array())
+
     writer.newline()
+    if message.has_attr("ifdef"):
+        writer.ifdef(message.attributes["ifdef"][0])
     parent_scope = writer.function(function_name,
                                    "uint8_t *",
                                    "uint8_t *message_start, uint8_t *message_end, int minor, size_t *size, message_destructor_t *free_message", True)
     parent_scope.variable_def("SPICE_GNUC_UNUSED uint8_t *", "pos");
     parent_scope.variable_def("uint8_t *", "start = message_start");
     parent_scope.variable_def("uint8_t *", "data = NULL");
-    parent_scope.variable_def("size_t", "mem_size", "nw_size");
+    parent_scope.variable_def("size_t", "nw_size")
+    if want_mem_size:
+        parent_scope.variable_def("size_t", "mem_size")
     if not message.has_attr("nocopy"):
         parent_scope.variable_def("uint8_t *", "in", "end");
     num_pointers = message.get_num_pointers()
@@ -1026,7 +1037,8 @@ def write_msg_parser(writer, message):
 
     write_parser_helpers(writer)
 
-    write_validate_container(writer, None, message, "start", parent_scope, True, True, False)
+    write_validate_container(writer, None, message, "start", parent_scope, True,
+                             want_mem_size=want_mem_size, want_extra_size=False)
 
     writer.newline()
 
@@ -1047,9 +1059,11 @@ def write_msg_parser(writer, message):
         writer.assign("end", "data + %s" % (msg_sizeof))
         writer.assign("in", "start").newline()
 
-        dest = RootDemarshallingDestination(None, msg_type, msg_sizeof, "data")
-        dest.reuse_scope = parent_scope
-        write_container_parser(writer, message, dest)
+        # avoid defined and assigned but not used warnings of gcc 4.6.0+
+        if message.is_extra_size() or not message.is_fixed_nw_size() or message.get_fixed_nw_size() > 0:
+            dest = RootDemarshallingDestination(None, msg_type, msg_sizeof, "data")
+            dest.reuse_scope = parent_scope
+            write_container_parser(writer, message, dest)
 
         writer.newline()
         writer.statement("assert(in <= message_end)")
@@ -1072,6 +1086,9 @@ def write_msg_parser(writer, message):
         writer.statement("return NULL")
     writer.end_block()
 
+    if message.has_attr("ifdef"):
+        writer.endif(message.attributes["ifdef"][0])
+
     return function_name
 
 def write_channel_parser(writer, channel, server):
@@ -1100,6 +1117,8 @@ def write_channel_parser(writer, channel, server):
     else:
         function_name = "parse_%s_msgc" % channel.name
     writer.newline()
+    if channel.has_attr("ifdef"):
+        writer.ifdef(channel.attributes["ifdef"][0])
     scope = writer.function(function_name,
                             "static uint8_t *",
                             "uint8_t *message_start, uint8_t *message_end, uint16_t message_type, int minor, size_t *size_out, message_destructor_t *free_message")
@@ -1129,6 +1148,8 @@ def write_channel_parser(writer, channel, server):
 
     writer.statement("return NULL")
     writer.end_block()
+    if channel.has_attr("ifdef"):
+        writer.endif(channel.attributes["ifdef"][0])
 
     return function_name
 
@@ -1145,13 +1166,16 @@ def write_get_channel_parser(writer, channel_parsers, max_channel, is_server):
 
     writer.write("static struct {spice_parse_channel_func_t func; unsigned int max_messages; } channels[%d] = " % (max_channel+1))
     writer.begin_block()
+    channel = None
     for i in range(0, max_channel + 1):
-        writer.write("{ ")
         if channel_parsers.has_key(i):
+            channel = channel_parsers[i][0]
+            if channel.has_attr("ifdef"):
+                writer.ifdef(channel.attributes["ifdef"][0])
+            writer.write("{ ")
             writer.write(channel_parsers[i][1])
             writer.write(", ")
 
-            channel = channel_parsers[i][0]
             max_msg = 0
             if is_server:
                 messages = channel.server_messages
@@ -1160,13 +1184,15 @@ def write_get_channel_parser(writer, channel_parsers, max_channel, is_server):
             for m in messages:
                 max_msg = max(max_msg, m.value)
             writer.write(max_msg)
+            writer.write("}")
         else:
-            writer.write("NULL, 0")
-        writer.write("}")
+            writer.write("{ NULL, 0 }")
 
         if i != max_channel:
             writer.write(",")
         writer.newline()
+        if channel and channel.has_attr("ifdef"):
+            writer.endif(channel.attributes["ifdef"][0])
     writer.end_block(semicolon = True)
 
     with writer.if_block("channel < %d" % (max_channel + 1)):
